# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""This module contains functions to parse the --help output of a program."""

import re
from collections import namedtuple


# Characters that should not be considered option chars
OPTION_BREAK_CHARS = [' ', '\t', '\n', ',', '|', '=', '[']

# Characters that delimit options
OPTION_DELIMITER_CHARS = [',', '|']

Unparsed = namedtuple('Unparsed', ['text'])
OptionWithMetavar = namedtuple('OptionWithMetavar', ['option', 'metavar', 'optional'])
OptionsWithDescription = namedtuple('OptionsWithDescription', ['options', 'description'])


class CharStream:
    """A utility class for sequentially reading characters from a string."""

    def __init__(self, string, pos = 0):
        self.string = string
        self.len = len(string)
        self.pos = pos

    def peek(self, relative_pos = 0):
        """
        Returns the character at the current position plus an optional relative
        offset without advancing the stream. Returns None if the position is
        out of bounds.
        """
        try:
            return self.string[self.pos + relative_pos]
        except IndexError:
            return None

    def peek_str(self, length):
        """
        Returns a substring of the specified length starting from the current
        position without advancing the stream.
        """
        return self.string[self.pos:self.pos + length]

    def get(self):
        """Returns the current character and advances the position by one."""
        c = self.string[self.pos]
        self.pos += 1
        return c

    def is_space(self):
        """Checks if the current character is a space or a tab."""
        return self.peek() in (' ', '\t')

    def is_end(self):
        """Checks if the current position has reached the end of the string."""
        return self.pos >= self.len

    def copy(self):
        """Creates and returns a new CharStream object at the current position."""
        return CharStream(self.string, self.pos)

    def __repr__(self):
        line = ''
        i = self.pos
        while i < self.len:
            if self.string[i] == '\n':
                break
            line += self.string[i]
            i += 1
        return f"CharStream({line!r})"


def eat_line(stream):
    '''Read the remainling line and return it (including the newline character).'''
    content = ''
    while not stream.is_end():
        char = stream.get()
        content += char
        if char == '\n':
            break
    return content


def eat_space(stream):
    '''Read spaces and tabs and return it.'''
    content = ''
    while stream.is_space():
        content += stream.get()
    return content


def parse_option_string(stream):
    '''Read an option string and return it.

    All chars except OPTION_BREAK_CHARS are considered valid option chars.

    Example option strings: --help, -h

    If the resulting option string is '-' or '--', it is not considered an option.
    '''
    option = ''
    p = stream.copy()

    eat_space(p)

    if p.peek() != '-':
        return None

    while not p.is_end() and p.peek() not in OPTION_BREAK_CHARS:
        option += p.get()

    if option in ('-', '--'):
        return None

    stream.pos = p.pos
    return option


def parse_bracket(stream):
    '''Read and return a bracketed expression.

    Bracketed expressions are:
        <foo bar>
        [foo bar]
        (foo bar)
        {foo bar}
    '''
    content = stream.peek()
    try:
        closing = {'<':'>', '[':']', '(':')', '{':'}'}[content]
    except KeyError:
        return None

    stream.get()
    while not stream.is_end():
        char = stream.get()
        content += char
        if char == closing:
            break

    return content


def parse_quoted_string(stream):
    '''Read and return a string.

    Strings are:
        'foo bar'
        "foo bar"

    Since it is unlikely that we encounter escape sequences in a description
    string of an option, we don't process any escape sequences.
    '''
    quote = stream.peek()
    if quote not in ('"', "'"):
        return None

    stream.get()
    content = quote
    while not stream.is_end():
        char = stream.get()
        content += char
        if char == quote:
            break

    return content


def parse_metavar(stream):
    '''Read and return a metavar.

    Everything until a tab, space or newline is considered a metavar.

    Special cases:
      - Bracketed expressions (e.g., '<foo bar>') and quoted strings (e.g., '"foo bar"')
        are handled, and the spaces within them are preserved.
      - The function supports metavars enclosed by `<`, `[`, `(`, `{`, as well as
        single (`'`) and double (`"`) quotes.

    Metavars are:
        foo_bar
        'foo bar'
        "foo bar"
        <foo bar>
    '''
    metavar = ''

    while not stream.is_end() and stream.peek() not in (' ', '\t', '\n'):
        if stream.peek() in ('<', '[', '(', '{'):
            metavar += parse_bracket(stream)
        elif stream.peek() in ('"', "'"):
            metavar += parse_quoted_string(stream)
        elif stream.peek() in OPTION_DELIMITER_CHARS:
            break
        else:
            metavar += stream.get()

    return metavar


def parse_trailing_description_line(stream):
    '''Reads and returns a trailing description line.

    A line is considered a trailing description line if it meets the following criteria:
      - It starts with whitespace (indicating continuation from a previous line).
      - It does not begin with a hyphen ('-'), which would indicate the start of a new option.
    '''
    p = stream.copy()

    if not p.is_space():
        return None

    space = eat_space(p)

    if p.peek() == '-' and len(space) < 10:
        return None

    content = eat_line(p)
    stream.pos = p.pos
    return content


def parse_description(stream):
    '''Reads and returns the description of an option.'''
    eat_space(stream)
    content = eat_line(stream)
    while True:
        line = parse_trailing_description_line(stream)
        if line:
            content += line
        else:
            break

    return content


def parse_option_with_metavar(stream):
    '''Read and return an option with its metavar (if any).

    Valid inputs are:
      --option=METAVAR
      --option[=METAVAR] (in this case, 'optional' is set to True)
      --option METAVAR

    Invalid inputs are:
      --option  METAVAR (notice two spaces)
    '''
    opt = parse_option_string(stream)
    metavar = None
    optional = False

    if opt:
        if stream.peek_str(2) == '[=':
            optional = True
            metavar = parse_metavar(stream)

        elif stream.peek() == '=':
            stream.get()
            metavar = parse_metavar(stream)

        # Two spaces after --option means the description follows
        elif stream.peek_str(2).isspace():
            return OptionWithMetavar(opt, metavar, optional)

        # An option delimiter cannot be a metavar
        elif parse_option_delimiter(stream.copy()):
            return OptionWithMetavar(opt, metavar, optional)

        elif not stream.is_end() and stream.is_space():
            stream.get()
            return OptionWithMetavar(opt, parse_metavar(stream), optional)

        return OptionWithMetavar(opt, metavar, optional)

    return None


def parse_option_delimiter(stream):
    '''Parse an option delimiter and return True if it was found, False otherwise.'''
    p = stream.copy()
    eat_space(p)
    if p.get() in (',', '|'):
        stream.pos = p.pos
        return True
    return False


def parse_options_with_description(stream):
    '''Parse options with description.'''
    options = []
    description = None

    while not stream.is_end():
        option = parse_option_with_metavar(stream)
        if option:
            options.append(option)
        else:
            break

        if not parse_option_delimiter(stream):
            break

    if not options:
        return None

    # if s.peek() == '\n' or s.peek_str(2) in ('  ', ' \n'):
    if stream.peek() in (' ', '\t', '\n'):
        description = parse_description(stream)

    return OptionsWithDescription(options, description)


def parse(stream):
    """Parses the stream and returns a list of options with descriptions or unparsed lines."""
    r = []

    while not stream.is_end():
        options = parse_options_with_description(stream)
        if options:
            r.append(options)
        else:
            line = eat_line(stream)
            r.append(Unparsed(line))

    return r


def get_program_name_from_help(string):
    """Extracts the program name from the help string."""
    m = re.match('usage:[\n\t ]+([^\n\t ]+)', string, re.I)
    if m:
        return m[1]
    return string.split()[0]



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""
This module provides functions for creating CommandLine objects from
Python's argparse.ArgumentParser.
"""

import argparse
from collections import OrderedDict

from . import file_loader
from . import utils
from .errors import CrazyError
from .cli import CommandLine, MutuallyExclusiveGroup

# We have to use implementation details of the argparse module...
# pylint: disable=protected-access
# pylint: disable=redefined-builtin


def range_to_complete(r):
    '''Convert a Python range object to a range complete format.'''

    start = r.start
    step = r.step
    end = r.stop - 1 if step > 0 else r.stop + 1

    if step == 1:
        return ('range', start, end)

    return ('range', start, end, step)


def get_complete(action):
    '''
    Get the `complete` attribute of `action` if it is set.
    Otherwise determine `complete` from action type.
    '''

    complete = action.get_complete()

    if isinstance(action, (
            argparse._HelpAction,
            argparse._VersionAction,
            argparse._StoreTrueAction,
            argparse._StoreFalseAction,
            argparse._StoreConstAction,
            argparse._AppendConstAction,
            argparse._CountAction)):

        if complete:
            utils.warn(f"Action has .complete() set but takes no arguments: {action}")

        return None

    if isinstance(action, (
            argparse._StoreAction,
            argparse._ExtendAction,
            argparse._AppendAction)):

        if complete:
            if action.choices:
                utils.warn(f"Action's choices overridden by .complete(): {action}")
            return complete

        if action.choices:
            if isinstance(action.choices, range):
                return range_to_complete(action.choices)
            return ['choices', action.choices]

        return ['none']

    if isinstance(action, argparse.BooleanOptionalAction):
        raise CrazyError("argparse.BooleanOptionalAction is not supported yet")

    raise CrazyError(f'Unknown action type: {action}')


def get_final(action):
    '''Return if an argparse action object is final.'''

    if isinstance(action, (argparse._HelpAction, argparse._VersionAction)):
        return True

    return action.get_final()


def argumentparser_to_commandline(parser, prog=None, description=None):
    '''Converts an ArgumentParser object to a CommandLine object.

    Args:
        parser (argparse.ArgumentParser):
            The ArgumentParser object to convert.

        prog (str, optional):
            The name of the program.
            Defaults to the program name of the parser.

        description (str, optional):
            The description of the program.
            Defaults to the description of the parser.

    Returns:
        CommandLine: A CommandLine object representing the converted parser.
    '''

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches

    if not description:
        description = parser.description

    if not prog:
        prog = parser.prog

    commandline = CommandLine(prog, help=description, aliases=parser.get_aliases())
    positional_number = 0

    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            subparsers = OrderedDict()
            for name, subparser in action.choices.items():
                subparsers[name] = {'parser': subparser, 'help': None}

            for subaction in action._get_subactions():
                subparsers[subaction.dest]['help'] = subaction.help

            subcommands = commandline.add_subcommands()
            for name, data in subparsers.items():
                subcmd = argumentparser_to_commandline(data['parser'], name, data['help'])
                subcommands.add_commandline_object(subcmd)
        elif not action.option_strings:
            if action.nargs in ('+', '*'):
                is_repeatable = True
            elif action.nargs in (1, None, '?'):
                is_repeatable = False
            else:
                is_repeatable = True
                utils.warn(f'Truncating nargs={action.nargs} of {action}')

            positional_number += 1
            commandline.add_positional(
                positional_number,
                metavar    = action.metavar or action.dest,
                complete   = get_complete(action),
                help       = action.help,
                repeatable = is_repeatable,
                when       = action.get_when()
            )
        else:
            complete = get_complete(action)
            optional_arg = False
            metavar = None
            help = None

            if action.nargs == '?':
                optional_arg = True
            elif action.nargs not in (None, 0, 1, True):
                utils.warn(f'Truncating nargs={action.nargs} of {action}')

            if complete:
                metavar = action.metavar or action.dest

            if action.help != argparse.SUPPRESS:
                help = action.help

            commandline.add_option(
                action.option_strings,
                metavar      = metavar,
                complete     = complete,
                help         = help,
                optional_arg = optional_arg,
                repeatable   = action.get_repeatable(),
                final        = get_final(action),
                hidden       = (action.help == argparse.SUPPRESS),
                when         = action.get_when()
            )

    group_counter = 0
    for group in parser._mutually_exclusive_groups:
        group_counter += 1
        group_name = f'group{group_counter}'

        exclusive_group = MutuallyExclusiveGroup(commandline, group_name)
        for action in group._group_actions:
            for option in commandline.get_options():
                for option_string in action.option_strings:
                    if option_string in option.option_strings:
                        exclusive_group.add_option(option)
                        break

    return commandline


def find_objects_by_type(module, types):
    '''Search a module for objects of a specific type.'''

    r = []

    for obj_name in dir(module):
        obj = getattr(module, obj_name)
        if isinstance(obj, types):
            r.append(obj)

    return r


def find_root_argument_parsers(module):
    '''Return a list of all ArgumentParser objects that have no parent.'''

    parsers = find_objects_by_type(module, argparse.ArgumentParser)
    actions = find_objects_by_type(module, argparse._SubParsersAction)

    for action in actions:
        for parser in action.choices.values():
            try:
                parsers.remove(parser)
            except ValueError:
                pass

    return parsers


def load_from_file(file, parser_variable=None, parser_blacklist=()):
    '''
    Load a Python file, search for the ArgumentParser object and convert
    it to a CommandLine object.
    '''

    # pylint: disable=broad-exception-caught

    try:
        module = file_loader.import_file(file)
    except Exception as e:
        utils.warn(e)
        utils.warn(f"Failed to load `{file}` using importlib, falling back to exec")
        module = file_loader.execute_file(file)

    if parser_variable is not None:
        try:
            parser = getattr(module, parser_variable)
        except AttributeError:
            raise CrazyError(f"No variable named `{parser_variable}` found in `{file}`") from None
    else:
        parsers = find_root_argument_parsers(module)
        for blacklisted in parser_blacklist:
            try:
                parsers.remove(blacklisted)
            except ValueError:
                pass
        if len(parsers) == 0:
            raise CrazyError(f"Could not find any ArgumentParser object in `{file}`")
        if len(parsers) > 1:
            raise CrazyError(f"Found too many ArgumentParser objects in `{file}`")
        parser = parsers[0]

    return argumentparser_to_commandline(parser)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Conditions for Fish.'''

import re

from . import cli
from . import when
from . import shell
from .errors import InternalError
from .type_utils import is_list_type


def escape_in_double(string):
    '''Escape special characters in a double quoted string.'''

    return re.sub(r'(["$`\\])', r'\\\1', string)


def make_condition_command(prefix, args):
    '''Make a condition command.

    In our completion scripts, we use the variables $query and $opts.

    We generate code like this:

        complete -n "$query '$opts' has_option --foo"

    This function generates such a condition command.

    We have to ensure proper escaping inside the double quotes.
    '''

    args_quoted = [shell.quote(arg) for arg in args]
    args_quoted = [escape_in_double(arg) for arg in args_quoted]
    return '%s %s' % (prefix, ' '.join(args_quoted))


def make_query(args):
    '''Make a query command.'''

    return make_condition_command("$query '$opts'", args)


class Condition:
    '''Base class for conditions.'''

    def query_code(self, ctxt):
        '''Returns code using the `query` function.'''
        raise NotImplementedError

    def unsafe_code(self, ctxt):
        '''Returns code using Fish's internal functions.'''
        raise NotImplementedError


class Not(Condition):
    '''Logical not.'''

    def __init__(self, conditional):
        self.conditional = conditional

    def query_code(self, ctxt):
        return 'not %s' % self.conditional.query_code(ctxt)

    def unsafe_code(self, ctxt):
        return 'not %s' % self.conditional.unsafe_code(ctxt)


class HasOption(Condition):
    '''Checks if an option is present on command line.'''

    def __init__(self, option_strings):
        if not is_list_type(option_strings):
            raise InternalError('option_strings: Invalid type')

        self.option_strings = option_strings

    def query_code(self, ctxt):
        ctxt.helpers.use_function('query', 'has_option')
        return make_query(['has_option'] + self.option_strings)

    def unsafe_code(self, _ctxt):
        args = []

        for opt in self.option_strings:
            if cli.is_long_option_string(opt):
                args += ['-l', opt.lstrip('-')]
            elif cli.is_short_option_string(opt):
                args += ['-s', opt[1]]
            else:
                args += ['-o', opt.lstrip('-')]

        return make_condition_command('__fish_seen_argument', args)


class HasHiddenOption(Condition):
    '''Checks if an incomplete option is present on command line.'''

    def __init__(self, option_strings):
        if not is_list_type(option_strings):
            raise InternalError('option_strings: Invalid type')

        self.option_strings = option_strings

    def query_code(self, ctxt):
        ctxt.helpers.use_function('query', 'has_option')
        ctxt.helpers.use_function('query', 'with_incomplete')
        options = self.option_strings
        return make_query(['has_option', 'WITH_INCOMPLETE'] + options)

    def unsafe_code(self, ctxt):
        return self.query_code(ctxt)


class OptionIs(Condition):
    '''Checks if an option has a specific value.'''

    def __init__(self, option_strings, values):
        if not is_list_type(option_strings):
            raise InternalError('option_strings: Invalid type')

        if not is_list_type(values):
            raise InternalError('values: Invalid type')

        self.option_strings = option_strings
        self.values = values

    def query_code(self, ctxt):
        ctxt.helpers.use_function('query', 'option_is')
        options = self.option_strings
        values = self.values
        return make_query(['option_is', *options, '--', *values])

    def unsafe_code(self, ctxt):
        return self.query_code(ctxt)


class PositionalNum(Condition):
    '''Check the number of positionals.'''

    TEST_OPERATORS = {
        '==': '-eq',
        '!=': '-ne',
        '<=': '-le',
        '>=': '-ge',
        '<':  '-lt',
        '>':  '-gt',
    }

    def __init__(self, operator, number):
        if operator not in ('==', '!=', '<', '<=', '>', '>='):
            raise InternalError(f'operator: Invalid value: {operator}')

        if not isinstance(number, int):
            raise InternalError(f'number: Invalid type: {number}')

        self.operator = operator
        self.number = number

    def query_code(self, ctxt):
        ctxt.helpers.use_function('query', 'num_of_positionals')
        op = self.TEST_OPERATORS[self.operator]
        return make_query(['num_of_positionals', op, str(self.number - 1)])

    def unsafe_code(self, _ctxt):
        op = self.TEST_OPERATORS[self.operator]
        number = self.number
        r = "test (__fish_number_of_cmd_args_wo_opts) %s %d" % (op, number)
        return r


class PositionalContains(Condition):
    '''Checks if a positional contains a specific value.'''

    def __init__(self, number, values):
        if not isinstance(number, int):
            raise InternalError(f'number: Invalid type: {number}')

        if not is_list_type(values):
            raise InternalError(f'values: Invalid type: {values}')

        self.number = number
        self.values = values

    def query_code(self, ctxt):
        ctxt.helpers.use_function('query', 'positional_contains')
        values = self.values
        number = str(self.number)
        return make_query(['positional_contains', number, *values])

    def unsafe_code(self, _ctxt):
        values = self.values
        return make_condition_command('__fish_seen_subcommand_from', values)


def replace_commands(tokens):
    '''Replace when objects by own condition objects.'''

    r = []

    for obj in tokens:
        if isinstance(obj, when.OptionIs):
            r.append(OptionIs(obj.options, obj.values))

        elif isinstance(obj, when.HasOption):
            r.append(HasOption(obj.options))

        elif isinstance(obj, str):
            r.append(obj)

        else:
            raise AssertionError("Not reached")

    return r


class Conditions(Condition):
    '''Holds a list of conditions.'''

    def __init__(self):
        self.conditions = []
        self.when = None

    def _optimized_conditions(self):
        positional_contains = []
        positional_num = []
        has_option = []
        other = []

        for condition in self.conditions:
            if isinstance(condition, PositionalContains):
                positional_contains.append(condition)
            elif isinstance(condition, PositionalNum):
                positional_num.append(condition)
            elif isinstance(condition, (HasHiddenOption, HasOption)):
                has_option.append(condition)
            else:
                other.append(condition)

        # positional_num is fastest, followed by positional_contains
        r = []
        r.extend(positional_num)
        r.extend(positional_contains)
        r.extend(has_option)
        r.extend(other)
        return r

    def _get_tokens(self):
        r = []

        for condition in self._optimized_conditions():
            r.append(condition)
            r.append('&&')

        if self.when:
            if r and when.needs_braces(self.when):
                r += ['(', *self.when, ')']
            else:
                r += self.when

        if r and r[-1] == '&&':
            r.pop(-1)

        return r

    def query_code(self, ctxt):
        r = []
        for obj in self._get_tokens():
            if obj == '(':
                r.append('begin')
            elif obj == ')':
                r.append(';end')
            elif obj in ('!', '&&', '||'):
                r.append(obj)
            else:
                r.append(obj.query_code(ctxt))

        if not r:
            return None

        return '"%s"' % ' '.join(r)

    def unsafe_code(self, ctxt):
        r = []
        for obj in self._get_tokens():
            if obj == '(':
                r.append('begin')
            elif obj == ')':
                r.append(';end')
            elif obj in ('!', '&&', '||'):
                r.append(obj)
            else:
                r.append(obj.unsafe_code(ctxt))

        if not r:
            return None

        return '"%s"' % ' '.join(r)

    def add(self, condition):
        '''Add a condition.'''

        self.conditions.append(condition)

    def extend(self, conditions):
        '''Add many conditions.'''

        self.conditions.extend(conditions)

    def add_when(self, tokens):
        '''Add when objects.'''

        if not tokens:
            return

        self.when = replace_commands(tokens)



import os
import sys
import subprocess

from terminal_base import TerminalBase

def run(args, env=None):
    result = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env=env)

    if result.returncode != 0:
        raise Exception("Command %r failed: %s" % (args, result.stderr))

    return result.stdout

try:
    run(['sh', '-c', 'which tmux'])
except Exception:
    print('Program `tmux` not found.')
    sys.exit(2)

class TmuxTerminal(TerminalBase):
    def __init__(self, session):
        self.session = session

    def _run(self, args):
        env = os.environ.copy()
        env.pop('TMUX', None)
        return run(['tmux'] + args, env)

    def start(self, command_args=(), env=None):
        tmux_env_args = []

        if env:
            for var, value in env.items():
                tmux_env_args.append(f'-e{var}={value}')

        cmd = ['new-session', '-d', '-s', self.session]
        cmd.extend(tmux_env_args)
        cmd.extend(command_args)
        self._run(cmd)

    def stop(self):
        self._run(['kill-session', '-t', self.session])

    def resize_window(self, x, y):
        self._run(['resize-window', '-t', self.session, '-x', str(x), '-y', str(y)])

    def send(self, line):
        self._run(['send-keys', '-t', self.session, line])

    def send_line(self, line):
        self._run(['send-keys', '-t', self.session, f'{line}\n'])

    def send_tab(self):
        self._run(['send-keys', '-t', self.session, 'Tab'])

    def send_ctrl(self, key):
        self._run(['send-keys', '-t', self.session, f'C-{key}'])

    def get_output(self):
        return self._run(['capture-pane', '-t', self.session, '-p'])



#!/usr/bin/env python3

import os
import re
import sys
import shlex
import argparse

# =============================================================================
# Constants
# =============================================================================

PROGRAM_PATHS = ['/bin', '/sbin', '/usr/bin', '/usr/sbin']

BASH_COMPLETION_PATHS = [
    '/etc/bash_completion.d',
    '/usr/share/bash-completion/completions',
]

FISH_COMPLETION_PATHS = [
    '/usr/share/fish/completions',
    '/usr/share/fish/vendor_completions.d',
]

ZSH_COMPLETION_PATHS = [
    '/usr/share/zsh/site-functions',
    '/usr/share/zsh/vendor-completions',
    '/usr/share/zsh/functions/Completion/Linux',
    '/usr/share/zsh/functions/Completion/Unix',
    '/usr/share/zsh/functions/Completion/Base',
    '/usr/share/zsh/functions/Completion/Zsh',
    '/usr/share/zsh/functions/Completion/X',
]

# =============================================================================
# Functions
# =============================================================================

def find_programs(path):
    try:
        return os.listdir(path)
    except FileNotFoundError:
        pass

def get_all_files_in_directories(directories):
    r = []

    for directory in directories:
        try:
            for file in os.scandir(directory):
                r.append(file.path)
        except FileNotFoundError:
            continue

    return r

def process_file(filename, callback):
    try:
        with open(filename, 'r', encoding='UTF-8') as fh:
            content = fh.read()
        return callback(filename, content)
    except FileNotFoundError:
        return None
    except UnicodeDecodeError:
        return None

def process_files(files, callback):
    r = []

    for file in files:
        with open(file, 'r', encoding='UTF-8') as fh:
            content = fh.read()
            result = callback(file, content)

            if result:
                r.append(result)

    return r

def parse_bash_complete_command(line):
    try:
        split = shlex.split(line)
    except Exception:
        return []

    i = 1
    have_F = False
    programs = []

    while i < len(split):
        arg = split[i]
        if arg == '-F':
            have_F = True
            i += 1
        elif arg == '-o':
            i += 1
        elif arg == '-X':
            i += 1
        elif arg.startswith('-'):
            pass
        else:
            programs.append(arg)
        i += 1

    if have_F:
        return programs
    else:
        return []

def parse_fish_complete_command(line):
    try:
        split = shlex.split(line)
    except Exception:
        return []

    i = 1
    while i < len(split):
        arg = split[i]
        if arg == '-c':
            return [split[i + 1]]
        elif arg.startswith('-c'):
            return [arg[2:]]
        i += 1

    return []

def parse_zsh_compdef(line):
    split = line.split()
    return split[1:]

def get_bash_completions():
    def callback(filename, content):
        r = []
        content = content.replace('\\\n', '')

        matches = re.findall('complete [^;&|\n]+', content)
        for m in matches:
            programs = parse_bash_complete_command(m)
            r.extend(programs)

        if not r:
            r = [os.path.basename(filename).lstrip('_')]

        return r

    files = get_all_files_in_directories(BASH_COMPLETION_PATHS)
    results = process_files(files, callback)

    r = set()
    for result in results:
        r.update(result)
    return sorted(r)

def get_fish_completions():
    def callback(filename, content):
        r = set()
        content = content.replace('\\\n', '')

        matches = re.findall('complete [^\n]+', content)
        for m in matches:
            programs = parse_fish_complete_command(m)
            r.update(programs)

        return r

    files = get_all_files_in_directories(FISH_COMPLETION_PATHS)
    results = process_files(files, callback)

    r = set()
    for result in results:
        r.update(result)
    return sorted(r)

def get_zsh_completions():
    def callback(filename, content):
        content = content.replace('\\\n', '')

        match = re.search('#compdef [^\n]+', content)
        if match:
            programs = parse_zsh_compdef(match[0])
            return programs

        return []

    files = get_all_files_in_directories(ZSH_COMPLETION_PATHS)
    results = process_files(files, callback)

    r = set()
    for result in results:
        r.update(result)
    return sorted(r)

# =============================================================================
# Classes
# =============================================================================

class Program:
    def __init__(self, program, has_bash, has_fish, has_zsh):
        self.program = program
        self.has_bash = has_bash
        self.has_fish = has_fish
        self.has_zsh = has_zsh

    def print(self):
        print('%-40s %-5s %-5s %-5s' % (
            self.program,
            self.has_bash,
            self.has_fish,
            self.has_zsh))

class Filter:
    def __init__(self, s):
        self.filters = {}

        for shell_filter in s.split(','):
            try:
                shell, filter_ = shell_filter.split('=')
            except ValueError:
                raise Exception('Invalid format. Format has to be SHELL=BOOL') from None

            if shell not in ['bash', 'fish', 'zsh']:
                raise Exception(f'Invalid shell: {shell}')

            if filter_ not in ['True', 'False']:
                raise Exception(f'Invalid filter: {filter_}')

            if shell in self.filters:
                raise Exception(f'Double shell key: {shell}')

            self.filters[shell] = {'True': True, 'False': False}[filter_]

    def accept(self, program):
        for shell, filter_ in self.filters.items():
            if shell == 'bash':
                if program.has_bash != filter_:
                    return False
            elif shell == 'fish':
                if program.has_fish != filter_:
                    return False
            elif shell == 'zsh':
                if program.has_zsh != filter_:
                    return False

        return True

# =============================================================================
# Command line arguments
# =============================================================================

p = argparse.ArgumentParser()
p.add_argument('-f', '--filter', metavar='SHELL=True|False',
    type=Filter, action='append',
    help='Filter results. Example: -f bash=True,zsh=False')
opts = p.parse_args()

# =============================================================================
# Collect data
# =============================================================================

bash_completions = get_bash_completions()
fish_completions = get_fish_completions()
zsh_completions = get_zsh_completions()

programs = set()
for path in PROGRAM_PATHS:
    programs.update(find_programs(path))
programs = sorted(programs)

# =============================================================================
# Filter and print results
# =============================================================================

for program in programs:
    program = Program(program,
        (program in bash_completions),
        (program in fish_completions),
        (program in zsh_completions))

    if opts.filter:
        for filter_ in opts.filter:
            if filter_.accept(program):
                program.print()
                break
    else:
        program.print()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Contains code for preprocessing text.'''


def preprocess(string, defines):
    '''Simple preprocessor function with #ifdef, #else, and #endif support.'''

    defines = set(defines)
    output = []
    stack = []  # stack of booleans: is this block currently active?

    lines = string.splitlines(keepends=True)

    for line in lines:
        stripped_line = line.lstrip()

        if stripped_line.startswith("#ifdef"):
            define = stripped_line.split()[1]
            active = define in defines
            stack.append(active)

        elif stripped_line.startswith("#else"):
            if not stack:
                raise SyntaxError("#else without #ifdef")
            # flip only the current (top of stack)
            stack[-1] = not stack[-1]

        elif stripped_line.startswith("#endif"):
            if not stack:
                raise SyntaxError("#endif without #ifdef")
            stack.pop()

        else:
            # output only if all enclosing blocks are active
            if not stack or all(stack):
                output.append(line)

    if stack:
        raise SyntaxError("Unclosed #ifdef")

    return ''.join(output)


def _test():
    string = """\
Lorem Ipsum
#ifdef DEBUG
    This is debug code.
    #ifdef FOO
    This is foo code.
    #else
    This is foo else code.
    #endif
#endif
Dolor Sit Amet\
"""
    defines = ["DEBUG"]
    result = preprocess(string, defines)
    print(result)


if __name__ == '__main__':
    _test()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for generating wrapper.'''

from . import cli
from . import algo
from . import preprocessor
from .str_utils import replace_many

_CODE = r'''
%WRAPPER_FUNC%() {
  %ORIGINAL_COMPLETION_FUNC%

  local i=0 del=0 adjust=0 delete=() new_words=()

  for ((i=1; i <= ${#words[@]}; ++i)); do
    local w="${words[$i]}"

    if false; then
      true
#ifdef long_opts_arg
    elif [[ "$w" == %LONG_OPTS_ARG_PATTERN1% ]]; then
      delete+=($i);
    elif [[ "$w" == %LONG_OPTS_ARG_PATTERN2% ]]; then
      delete+=($i $((++i)));
#endif
#ifdef long_opts_optional
    elif [[ "$w" == %LONG_OPTS_OPTIONAL_PATTERN% ]]; then
      delete+=($i);
#endif
#ifdef long_opts_flag
    elif [[ "$w" == %LONG_OPTS_FLAG_PATTERN% ]]; then
      delete+=($i);
#endif
#ifdef short_opts_arg
    elif [[ "$w" =~ '%SHORT_OPTS_ARG_REGEX1%' ]]; then
      delete+=($i $((++i)));
    elif [[ "$w" =~ '%SHORT_OPTS_ARG_REGEX2%' ]]; then
      delete+=($i);
#endif
#ifdef short_opts_optional
    elif [[ "$w" =~ '%SHORT_OPTS_OPTIONAL_REGEX%' ]]; then
      delete+=($i);
#endif
#ifdef short_opts_flag
    elif [[ "$w" =~ '%SHORT_OPTS_FLAG_REGEX%' ]]; then
      delete+=($i);
#endif
    fi
  done

  for del in "${delete[@]}"; do
    (( del < CURRENT )) && (( ++adjust ))
  done

  (( CURRENT -= adjust ))

  for ((i=1; i <= ${#words[@]}; ++i)); do
    if [[ " ${delete[@]} " != *" $i "* ]]; then
      new_words+=("${words[$i]}")
    fi
  done

  words=("${new_words[@]}")
  words[1]='%WRAPS%'
  service=%WRAPS%
  (( $+_comps[%WRAPS%] )) && $_comps[%WRAPS%]
}
'''


def _make_long_opts_pattern(opts, arg_type):
    if not opts:
        return ''

    long, old = algo.partition(opts, cli.is_long_option_string)

    old = [o.lstrip('-') for o in old]
    long = [o.lstrip('-') for o in long]
    r = ''

    if long and old:
        r = '-(%s|-(%s))' % ('|'.join(old), '|'.join(long))
    elif long:
        r = '--(%s)' % '|'.join(long)
    elif old:
        r = '-(%s)' % '|'.join(old)

    if arg_type == 'arg':
        r += '=*'
    elif arg_type == 'optional':
        r += '(|=*)'

    return r


def _make_short_opts_flag_regex(flag_opts):
    if not flag_opts:
        return ''

    return '^-[%s]+$' % flag_opts


def _make_short_opts_arg_regex(flag_opts, arg_opts, arg_type):
    if not arg_opts:
        return ''

    if flag_opts:
        r = '-[%s]*[%s]' % (flag_opts, arg_opts)
    else:
        r = '-[%s]' % arg_opts

    if arg_type == 'arg':
        r += '.+'
    elif arg_type == 'optional':
        r += '.*'

    return f'^{r}$'


def generate_wrapper(ctxt, commandline):
    '''Generate code for wrapping a foreign command.'''

    make_completion_funcname = ctxt.helpers.make_completion_funcname
    completion_funcname = make_completion_funcname(commandline)
    wrapper_funcname = make_completion_funcname(commandline, '__wrapper')

    if not commandline.wraps:
        return (completion_funcname, None)

    long_opts_arg = []
    long_opts_flag = []
    long_opts_optional = []

    short_opts_arg = []
    short_opts_flag = []
    short_opts_optional = []

    for option in commandline.get_options():
        if option.has_required_arg():
            long_opts_arg += option.get_long_option_strings()
            long_opts_arg += option.get_old_option_strings()
            short_opts_arg += option.get_short_option_strings()
        elif option.has_optional_arg():
            long_opts_optional += option.get_long_option_strings()
            long_opts_optional += option.get_old_option_strings()
            short_opts_optional += option.get_short_option_strings()
        else:
            long_opts_flag += option.get_long_option_strings()
            long_opts_flag += option.get_old_option_strings()
            short_opts_flag += option.get_short_option_strings()

    short_opts_flag = ''.join(o[1] for o in short_opts_flag)
    short_opts_arg = ''.join(o[1] for o in short_opts_arg)
    short_opts_optional = ''.join(o[1] for o in short_opts_optional)

    s = _CODE.strip()
    s = replace_many(s, [
        ('%ORIGINAL_COMPLETION_FUNC%',
         completion_funcname),

        ('%WRAPPER_FUNC%',
         wrapper_funcname),

        ('%WRAPS%',
         commandline.wraps),

        ('%LONG_OPTS_FLAG_PATTERN%',
         _make_long_opts_pattern(long_opts_flag, None)),

        ('%LONG_OPTS_ARG_PATTERN1%',
         _make_long_opts_pattern(long_opts_arg, 'arg')),

        ('%LONG_OPTS_ARG_PATTERN2%',
         _make_long_opts_pattern(long_opts_arg, None)),

        ('%LONG_OPTS_OPTIONAL_PATTERN%',
         _make_long_opts_pattern(long_opts_optional, 'optional')),

        ('%SHORT_OPTS_FLAG_REGEX%',
         _make_short_opts_flag_regex(short_opts_flag)),

        ('%SHORT_OPTS_ARG_REGEX1%',
         _make_short_opts_arg_regex(
             short_opts_flag, short_opts_arg, None)),

        ('%SHORT_OPTS_ARG_REGEX2%',
         _make_short_opts_arg_regex(
             short_opts_flag, short_opts_arg, 'arg')),

        ('%SHORT_OPTS_OPTIONAL_REGEX%',
         _make_short_opts_arg_regex(
             short_opts_flag, short_opts_optional, 'optional'))
    ])

    defines = []
    if long_opts_flag:
        defines.append('long_opts_flag')
    if long_opts_arg:
        defines.append('long_opts_arg')
    if long_opts_optional:
        defines.append('long_opts_optional')
    if short_opts_flag:
        defines.append('short_opts_flag')
    if short_opts_arg:
        defines.append('short_opts_arg')
    if short_opts_optional:
        defines.append('short_opts_optional')

    s = preprocessor.preprocess(s, defines)

    return (wrapper_funcname, s)



#!/usr/bin/python3

import os
import pprint
from collections import defaultdict

import yaml
from pygments import highlight
from pygments.lexers import YamlLexer
from pygments.lexers import MarkdownLexer
from pygments.formatters import TerminalFormatter

os.chdir(os.path.dirname(os.path.abspath(__file__)))

COMMANDS_INFILE         = 'commands.yaml'
COMMANDS_OUTFILE        = 'commands.md'
OPTIONS_INFILE          = 'options.yaml'
OPTIONS_OUTFILE         = 'options.md'
DOCUMENTATION_INFILE    = 'documentation.md.in'
DOCUMENTATION_OUTFILE   = 'documentation.md'
MANUAL_INFILE           = 'manual.in.py'
MANUAL_OUTFILE          = '../crazy_complete/manual.py'

OPTIONS                 = []
COMMANDS                = []
BY_CATEGORY             = defaultdict(list)

def escape_underscore(s):
    return s.replace('_', '\\_')

class Command:
    def __init__(self, dictionary):
        self.command = None
        self.category = None
        self.synopsis = None
        self.short = None
        self.long = None
        self.notes = []
        self.definition = None
        self.output = None
        self.also = None
        self.implemented = None

        for key, value in dictionary.items():
            if key == 'command':
                assert isinstance(value, str), key
                self.command = value
            elif key == 'synopsis':
                assert isinstance(value, str), key
                self.synopsis = value
            elif key == 'short':
                assert isinstance(value, str), key
                self.short = value
            elif key == 'long':
                assert isinstance(value, str), key
                self.long = value
            elif key == 'definition':
                assert isinstance(value, str), key
                self.definition = value
            elif key == 'category':
                assert isinstance(value, str), key
                self.category = value
            elif key == 'output':
                assert isinstance(value, str), key
                self.output = value
            elif key == 'notes':
                assert isinstance(value, list), key
                self.notes = value
            elif key == 'also':
                assert isinstance(value, dict), key
                self.also = value
            elif key == 'implemented':
                assert isinstance(value, list), key
                self.implemented = value
            else:
                raise Exception(f"Unknown key: {key}")

        if self.command is None:
            raise Exception("No command")

        if self.category is None:
            raise Exception(f"No category: {self.command}")

        if self.category not in ('meta', 'custom', 'basic', 'bonus'):
            raise Exception(f"No such category: {self.category}")

        if self.short is None:
            raise Exception(f"No short: {self.command}")

        if self.short.endswith('.'):
            raise Exception(f"short ends with '.': {self.command}")

        if self.definition is None:
            raise Exception(f"No definition: {self.command}")

        yaml.safe_load(self.definition)

        if self.synopsis is None:
            self.synopsis = f'["{self.command}"]'

        if self.output is None:
            raise Exception(f"No output: {self.command}")

        if '<TAB>' not in self.output:
            raise Exception(f"No <TAB> in output: {self.command}")

        if self.implemented:
            for implemented in self.implemented:
                if not implemented in ('Bash', 'Fish', 'Zsh'):
                    raise Exception(f"Implemented: Wrong shell {implemented}")

class Option:
    def __init__(self, dictionary):
        self.options = None
        self.metavar = None
        self.choices = None
        self.default = None
        self.short = None
        self.long = None
        self.optional = False

        for key, value in dictionary.items():
            if key == 'options':
                assert isinstance(value, list), key
                self.options = value
            elif key == 'short':
                assert isinstance(value, str), key
                self.short = value
            elif key == 'long':
                assert isinstance(value, str), key
                self.long = value
            elif key == 'metavar':
                assert isinstance(value, str), key
                self.metavar = value
            elif key == 'default':
                assert isinstance(value, str), key
                self.default= value
            elif key == 'choices':
                assert isinstance(value, (list, dict)), key
                self.choices = value
            elif key == 'optional':
                assert isinstance(value, bool), key
                self.optional = value
            else:
                raise Exception(f"Unknown key: {key}")

        if self.options is None:
            raise Exception("No options")

        if self.short is None:
            raise Exception(f"No short: {self.options}")


def make_markup_table(header, rows):
    # Ensure everything is string
    header = [str(h) for h in header]
    rows = [[str(c) for c in row] for row in rows]

    # Determine column widths (max of header + rows)
    num_cols = len(header)
    col_widths = [
        max(len(header[i]), max((len(row[i]) for row in rows), default=0))
        for i in range(num_cols)
    ]

    # Helper to format a row
    def fmt_row(row):
        return "| " + " | ".join(
            row[i].ljust(col_widths[i]) for i in range(num_cols)
        ) + " |"

    # Header separator
    separator = "| " + " | ".join("-" * col_widths[i] for i in range(num_cols)) + " |"

    # Build table
    table_lines = [fmt_row(header), separator]
    for row in rows:
        table_lines.append(fmt_row(row))

    return "\n".join(table_lines)


def make_english_listing(words):
    if len(words) == 1:
        return words[0]
    elif len(words) == 2:
        return f'{words[0]} and {words[1]}'
    elif len(words) == 3:
        return f'{words[0]}, {words[1]} and {words[2]}'

    raise AssertionError("Not reached")


def make_implemented_sentence(shells):
    shells_formatted = [f'**{w}**' for w in sorted(shells)]
    r = f'This completer is currently only implemented in {make_english_listing(shells_formatted)}.'
    return r


def make_markup_link(text, link):
    return f'[{escape_underscore(text)}](#{link})'


def make_markup_summary_table(header, commands, pre_text=None):
    r = [f'### {escape_underscore(header)}']

    if pre_text:
        r += [pre_text]

    header = ['Command', 'Description']
    rows = []
    for c in commands:
        rows.append([make_markup_link(c.command, c.command), escape_underscore(c.short)])

    r += [make_markup_table(header, rows)]

    return '\n\n'.join(r)

def make_markup_command(command):
    r =  [f'### {escape_underscore(command.command)}']
    r += [f'> {command.short}']
    if command.long:
        r += [command.long]

    notes = list(command.notes)
    if command.implemented:
        notes += [make_implemented_sentence(command.implemented)]

    if notes:
        r += ['**NOTES**']
        r += [f'- {l}' for l in notes]

    r += [f'```yaml\n{command.definition.strip()}\n```']
    if command.output:
        r += [f'```\n{command.output.strip()}\n```']
    if command.also:
        r += ['**SEE ALSO**']
        r += [f'- {make_markup_link(cmd, cmd)}: {desc}' for cmd, desc in command.also.items()]

    return '\n\n'.join(r)


def make_markup(by_category):
    r = []
    r += [make_markup_summary_table('Meta commands', BY_CATEGORY['meta'])]
    r += [make_markup_summary_table('Built-in commands', BY_CATEGORY['basic'])]
    r += [make_markup_summary_table('User-defined commands', BY_CATEGORY['custom'])]
    r += [make_markup_summary_table('Bonus commands', BY_CATEGORY['bonus'])]

    for commands in by_category.values():
        for command in commands:
            r += [make_markup_command(command)]

    return '\n\n---\n\n'.join(r)


def make_option_markup(option):
    option_strings = '|'.join(option.options)

    r = f'**{option_strings}'
    if option.metavar:
        r += '='
    if option.optional:
        r += '['
    if option.metavar:
        r += option.metavar
    if option.optional:
        r += ']'
    r += '**'

    if option.choices:
        choices = ', '.join(option.choices)
        r += f' *({choices})*'

    r += f'\n\n> {option.short}'

    if option.long:
        r += f'\n\n{option.long.strip()}'

    if isinstance(option.choices, dict):
        r += '\n'
        for key, value in option.choices.items():
            r += f'\n- {key}: {value}'

    if option.default:
        r += f'\n\nThis option defaults to `{option.default}`.'

    return r

def options_markup(options):
    r = []
    for option in options:
        r.append(make_option_markup(option))

    return '\n\n---\n\n'.join(r)


def make_python_commands(commands):
    data = []

    for command in commands:
        def_colored = highlight(command.definition, YamlLexer(), TerminalFormatter())

        if command.long:
            long_colored = highlight(command.long, MarkdownLexer(), TerminalFormatter())
        else:
            long_colored = None

        data.append({
            'command':              command.command,
            'category':             command.category,
            'short':                command.short,
            'long':                 command.long,
            'long_colored':         long_colored,
            'notes':                command.notes,
            'implemented':          command.implemented,
            'definition':           command.definition,
            'definition_colored':   def_colored,
            'output':               command.output,
            'also':                 command.also
        })

    return 'COMMANDS = ' + pprint.pformat(data)

# =============================================================================
# Read commands and options
# =============================================================================

with open(COMMANDS_INFILE, 'r', encoding='utf-8') as fh:
    for command_def in yaml.safe_load_all(fh):
        COMMANDS.append(Command(command_def))

with open(OPTIONS_INFILE, 'r', encoding='utf-8') as fh:
    for option_def in yaml.safe_load_all(fh):
        OPTIONS.append(Option(option_def))

COMMANDS = sorted(COMMANDS, key=lambda c: c.command)

# =============================================================================
# Generate command and option files
# =============================================================================

for command in COMMANDS:
    BY_CATEGORY[command.category].append(command)

commands_markup = make_markup(BY_CATEGORY)
options_markup = options_markup(OPTIONS)

with open(COMMANDS_OUTFILE, 'w', encoding='utf-8') as fh:
    fh.write(commands_markup)

with open(OPTIONS_OUTFILE, 'w', encoding='utf-8') as fh:
    fh.write(options_markup)

with open(DOCUMENTATION_INFILE, 'r', encoding='utf-8') as fh:
    content = fh.read()

content = content.replace('%COMMANDS%', commands_markup)
content = content.replace('%OPTIONS%', options_markup)

with open(DOCUMENTATION_OUTFILE, 'w', encoding='utf-8') as fh:
    fh.write(content)

# =============================================================================
# Generate python manual data
# =============================================================================

with open(MANUAL_INFILE, 'r', encoding='utf-8') as fh:
    content = fh.read()

commands_python = make_python_commands(COMMANDS)
content = content.replace('%COMMANDS%', commands_python)

with open(MANUAL_OUTFILE, 'w', encoding='utf-8') as fh:
    fh.write(content)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains functions for transforming bash glob patterns.'''


from .string_stream import StringStream


class PatternBase:
    '''Base class for glob patterns.'''

    def to_regex(self):
        '''Make a regular expression.'''
        raise NotImplementedError

    def to_zsh_glob(self):
        '''Make a Zsh glob.'''
        raise NotImplementedError


class Literal(PatternBase):
    '''Holds a literal string.'''

    REGEX_ESCAPE_CHARS = '.?*+|{}[]()'

    def __init__(self, string):
        self.string = string

    def __str__(self):
        return self.string

    def __repr__(self):
        return self.string

    def to_regex(self):
        string = self.string

        for c in Literal.REGEX_ESCAPE_CHARS:
            string = string.replace(c, f'\\{c}')

        return string

    def to_zsh_glob(self):
        string = ''

        for c in self.string:
            if c in ('[', ']', '?', '*', '"', ' ', '\\'):
                string += f"'{c}'"
            elif c == "'":
                string += f'"{c}"'
            else:
                string += c

        return string


class Star(PatternBase):
    '''Represents a star glob (*).'''

    def to_regex(self):
        return '.*'

    def to_zsh_glob(self):
        return '*'


class Question(PatternBase):
    '''Represents a question glob (?).'''

    def to_regex(self):
        return '.'

    def to_zsh_glob(self):
        return '?'


class CharClass(PatternBase):
    '''Represents a character class.'''

    REGEX_ESCAPE_CHARS = r'^\.[]'

    def __init__(self, chars, negated):
        self.chars = chars
        self.negated = negated

    def to_regex(self):
        chars = self.chars

        for c in CharClass.REGEX_ESCAPE_CHARS:
            chars = chars.replace(c, f'\\{c}')

        if self.negated:
            return f'[^{chars}]'

        return f'[{chars}]'

    def to_zsh_glob(self):
        chars = ''

        for c in self.chars:
            if c in ('[', ']', '!'):
                chars += f"'{c}'"
            else:
                chars += c

        if self.negated:
            return f'[!{chars}]'

        return f'[{chars}]'


class ExtGlob(PatternBase):
    '''Holds an extglob: @(), *(), ?(), +(), !().'''

    def __init__(self, type_, alternatives):
        self.type = type_
        self.alternatives = alternatives

    def to_regex(self):
        r = []

        for alternative in self.alternatives:
            r.append(alternative.to_regex())

        p = '(%s)' % '|'.join(r)

        if self.type == '@':
            return p

        if self.type == '*':
            return p + '*'

        if self.type == '+':
            return p + '+'

        if self.type == '?':
            return p + '?'

        if self.type == '!':
            raise ValueError('Negated glob !(...) not supported in regex')

        raise AssertionError("Not reached")

    def to_zsh_glob(self):
        r = []

        for alternative in self.alternatives:
            r.append(alternative.to_zsh_glob())

        p = '(%s)' % '|'.join(r)

        if self.type == '@':
            return p

        if self.type in ('*', '+', '?', '!'):
            raise ValueError(f'Extended glob of form {self.type}(...) not supported in zsh glob')

        raise AssertionError("Not reached")


class PatternList(PatternBase):
    '''Holds a list of patterns.'''

    def __init__(self):
        self.tokens = []

    def to_regex(self):
        return ''.join(t.to_regex() for t in self.tokens)

    def to_zsh_glob(self):
        return ''.join(t.to_zsh_glob() for t in self.tokens)

    def append(self, token):
        '''Add a token.'''
        self.tokens.append(token)


EXTGLOB_TOKENS = ('@(', '*(', '+(', '?(', '!(', '+(')
EXTGLOB_TOKENS_PRE = ('@', '*', '+', '?', '!', '+')


class GlobLexer(StringStream):
    '''Class for splitting a pattern into tokens.'''

    def __init__(self, pattern):
        super().__init__(pattern)
        self.tokens = []

    def parse(self):
        '''Split string into tokens.'''
        tokens = []

        while True:
            try:
                tokens.append(self._parse_token())
            except IndexError:
                return tokens

    def _parse_token(self):
        c = self.get()

        if c in EXTGLOB_TOKENS_PRE and self.peek() == '(':
            return f'{c}{self.get()}'

        if c in ('[', ']', ')', '?', '!', '*', '|'):
            return c

        if c == '"':
            return Literal(self.parse_shell_double_quote(in_quotes=True))

        if c == "'":
            return Literal(self.parse_shell_single_quote(in_quotes=True))

        literal = c
        while True:
            c = self.peek()
            if c is None:
                break
            if c in EXTGLOB_TOKENS_PRE and self.peek(1) == '(':
                break
            if c in ('*', '?', '[', ']', ')', '|', '"', "'"):
                break
            literal += c
            self.get()

        return Literal(literal)


class GlobParser:
    '''Parses tokens from GlobLexer.'''

    def __init__(self, tokens):
        self.tokens = tokens
        self.i = 0

    def get(self):
        '''Return a token and advance position.'''
        token = self.tokens[self.i]
        self.i += 1
        return token

    def peek(self):
        '''Return current token without advancing the position.'''
        try:
            return self.tokens[self.i]
        except IndexError:
            return None

    def have(self):
        '''Return True if there are any tokens left.'''
        return self.i < len(self.tokens)

    def parse(self):
        '''Parse tokens and return a PatternList object.'''
        root = PatternList()
        while self.have():
            root.append(self._parse_token())
        return root

    def _parse_char_class(self):
        chars = ''
        negated = False

        try:
            token = self.get()
        except IndexError as e:
            raise ValueError('Unclosed character class') from e

        if token == '!':
            negated = True
        elif token == ']':
            raise ValueError('Empty character class')
        else:
            chars += str(token)

        while True:
            try:
                token = self.get()
            except IndexError as e:
                raise ValueError('Unclosed character class') from e

            if token == ']':
                return CharClass(chars, negated)

            chars += str(token)

    def _parse_ext_glob(self, token):
        type_ = token[0]
        alternatives = [PatternList()]

        while True:
            token = self.peek()

            if token == ')':
                self.get()
                return ExtGlob(type_, alternatives)

            if token == '|':
                self.get()
                alternatives.append(PatternList())
            elif token is None:
                raise ValueError('Unclosed extglob')
            else:
                alternatives[-1].append(self._parse_token())

    def _parse_token(self):
        token = self.get()

        if token == '[':
            return self._parse_char_class()

        if token == '*':
            return Star()

        if token == '?':
            return Question()

        if token in EXTGLOB_TOKENS:
            return self._parse_ext_glob(token)

        return token


def bash_glob_to_regex(pattern):
    '''Transform a bash glob to regex.'''

    tokens = GlobLexer(pattern).parse()
    tokens = GlobParser(tokens).parse()
    return tokens.to_regex()


def bash_glob_to_zsh_glob(pattern):
    '''Transform a bash glob to zsh glob.'''

    tokens = GlobLexer(pattern).parse()
    tokens = GlobParser(tokens).parse()
    return tokens.to_zsh_glob()


def test():
    '''Tests.'''

    def do_test(num, bash_pattern, expected_regex, expected_zsh_glob):
        '''Do a test case.'''

        try:
            regex_result = bash_glob_to_regex(bash_pattern)
        except ValueError:
            regex_result = ValueError

        try:
            zsh_glob_result = bash_glob_to_zsh_glob(bash_pattern)
        except ValueError:
            zsh_glob_result = ValueError

        if regex_result != expected_regex:
            print('Test %d (regex):' % num)
            print('Having:   %s' % regex_result)
            print('Expected: %s' % expected_regex)
            raise AssertionError('Test failed')

        if zsh_glob_result != expected_zsh_glob:
            print('Test %d (zsh glob):' % num)
            print('Having:   %s' % zsh_glob_result)
            print('Expected: %s' % expected_zsh_glob)
            raise AssertionError('Test failed')

    tests = [
        # Literals
        (0,  r'a',                   r'a',                  r'a'),
        (1,  r'abc',                 r'abc',                r'abc'),
        (3,  r'"abc"',               r'abc',                r'abc'),
        (4,  r"'abc'",               r'abc',                r'abc'),
        (5,  r"'a'b'c'",             r'abc',                r'abc'),
        (6,  r'.',                   r'\.',                 r'.'),
        (7,  r'+',                   r'\+',                 r'+'),
        (8,  r'@',                   r'@',                  r'@'),
        (9,  r'"["',                 r'\[',                 r"'['"),
        (10, r"'['",                 r'\[',                 r"'['"),
        (11, r'"]"',                 r'\]',                 r"']'"),
        (12, r"']'",                 r'\]',                 r"']'"),
        (13, r"'*'",                 r'\*',                 r"'*'"),
        (14, r"'?'",                 r'\?',                 r"'?'"),

        # Simple globbing
        (15, r'*',                   r'.*',                 r'*'),
        (16, r'?',                   r'.',                  r'?'),
        (17, r'???',                 r'...',                r'???'),

        # Literals + simple globbing
        (18, r'*abc*',               r'.*abc.*',            r'*abc*'),
        (19, r'*.*',                 r'.*\..*',             r'*.*'),
        (20, r'*+*',                 r'.*\+.*',             r'*+*'),

        # Character classes
        (21, r'[abc]',               r'[abc]',              r'[abc]'),
        (22, r'[.]',                 r'[\.]',               r'[.]'),
        (23, r'[+]',                 r'[+]',                r'[+]'),
        (24, r'[}]',                 r'[}]',                r'[}]'),
        (25, r'[{]',                 r'[{]',                r'[{]'),
        (26, r'[*]',                 r'[*]',                r'[*]'),
        (27, r'[(]',                 r'[(]',                r'[(]'),
        (28, r'[)]',                 r'[)]',                r'[)]'),
        (29, r'["["]',               r'[\[]',               r"['[']"),
        (30, r'["]"]',               r'[\]]',               r"[']']"),

        # Extglob @
        (31, r'@(foo)',              r'(foo)',              r'(foo)'),
        (32, r'@(foo|bar)',          r'(foo|bar)',          r'(foo|bar)'),
        (33, r'@(|foo)',             r'(|foo)',             r'(|foo)'),

        # Extglob *
        (34, r'*(foo|bar)',          r'(foo|bar)*',         ValueError),
        (35, r'*(foo)',              r'(foo)*',             ValueError),
        (36, r'*(|foo)',             r'(|foo)*',            ValueError),

        # Extglob +
        (37, r'+(foo|bar)',          r'(foo|bar)+',         ValueError),
        (38, r'+(foo)',              r'(foo)+',             ValueError),
        (39, r'+(|foo)',             r'(|foo)+',            ValueError),

        # Extglob ?
        (40, r'?(foo|bar)',          r'(foo|bar)?',         ValueError),
        (41, r'?(foo)',              r'(foo)?',             ValueError),
        (42, r'?(|foo)',             r'(|foo)?',            ValueError),

        # Extglob !
        (43, r'!(foo|bar)',          ValueError,            ValueError),

        # Nested
        (44, r'@([abc]|foo)',        r'([abc]|foo)',        r'([abc]|foo)'),
        (45, r'@(@(foo|bar))',       r'((foo|bar))',        r'((foo|bar))'),
        (46, r'@(@("foo"|bar))',     r'((foo|bar))',        r'((foo|bar))'),
        (47, r'@(?|bar)',            r'(.|bar)',            r'(?|bar)'),
        (48, r'@(*baz|bar)',         r'(.*baz|bar)',        r'(*baz|bar)'),

        # Syntax errors
        (70, r'@(foo',               ValueError,            ValueError),
        (71, r'[abc',                ValueError,            ValueError),
        (71, r'"abc',                ValueError,            ValueError),
        (71, r"'abc",                ValueError,            ValueError),
    ]

    for case in tests:
        do_test(*case)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Exporting modules.'''

from . import errors
from . import cli, config, utils
from . import shell, bash, fish, zsh
from . import argparse_source, json_source, yaml_source
from . import extended_yaml_parser
from . import scheme_validator
from . import help_parser

__all__ = [
    'errors',
    'cli',
    'config',
    'utils',
    'shell',
    'bash',
    'fish',
    'zsh',
    'argparse_source',
    'json_source',
    'yaml_source',
    'extended_yaml_parser',
    'scheme_validator',
    'help_parser',
]



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''String utility functions.'''

import re
import subprocess

from .errors import CrazyError


_VALID_OPTION_STRING_RE = re.compile('-[^\\s,]+')
_VALID_VARIABLE_RE = re.compile('[a-zA-Z_][a-zA-Z0-9_]*')


def contains_space(string):
    '''Check if string contains space type characters.'''

    return (' ' in string or '\t' in string or '\n' in string)


def is_empty_or_whitespace(string):
    '''Check if string is empty or whitespace.'''

    return not string.strip()


def is_valid_extended_regex(string):
    '''Check if string is a valid extended regular expression.'''

    try:
        r = subprocess.run(
            ['grep', '-q', '-E', '--', string],
            input=b"",
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            timeout=1,
            check=False)

        return r.returncode != 2
    except (FileNotFoundError, PermissionError, subprocess.TimeoutExpired):
        return True


def validate_prog(string):
    '''Validate a program string.'''

    if is_empty_or_whitespace(string):
        raise CrazyError('value is empty')

    if string.startswith(' '):
        raise CrazyError('begins with space')

    if string.endswith(' '):
        raise CrazyError('ends with space')

    if '\t' in string:
        raise CrazyError('contains a tabulator')

    if '\n' in string:
        raise CrazyError('contains a newline')

    if '  ' in string:
        raise CrazyError('contains multiple spaces')


def indent(string, num_spaces):
    '''Indents each line in a string by a specified number of spaces,
    preserving empty lines.

    Args:
        string (str): The input string to be indented.
        num_spaces (int): The number of spaces to indent each line.

    Returns:
        str: The indented string.
    '''

    assert isinstance(string, str), \
        f"indent: string: expected str, got {string!r}"

    assert isinstance(num_spaces, int), \
        f"indent: num_spaces: expected int, got {num_spaces!r}"

    spaces = ' ' * num_spaces
    lines = string.split('\n')
    lines = [(spaces + line) if line.strip() else line for line in lines]
    return '\n'.join(lines)


def join_with_wrap(primary_sep, secondary_sep,
                   max_line_length, items, line_prefix=''):
    '''
    Join a list of strings with a primary separator, automatically wrapping
    to a new line (using the secondary separator) when the current line
    would exceed max_line_length.

    Args:
        primary_sep:
            String used to separate items on the same line.

        secondary_sep:
            String used to separate wrapped lines.

        max_line_length:
            Maximum number of characters per line.

        items:
            List of strings to join.

        line_prefix:
            Prefix each line with a prefix.

    Returns:
        str: Joined string with line wrapping
    '''

    lines = []
    current_items = []
    current_length = len(line_prefix)

    for item in items:
        sep_len = len(primary_sep) if current_items else 0
        new_length = current_length + sep_len + len(item)

        if current_items and new_length > max_line_length:
            # flush current line
            lines.append(line_prefix + primary_sep.join(current_items))
            current_items = [item]
            current_length = len(line_prefix) + len(item)
        else:
            current_items.append(item)
            current_length = new_length

    if current_items:
        lines.append(line_prefix + primary_sep.join(current_items))

    return secondary_sep.join(lines)


def is_valid_option_string(option_string):
    '''Check if `option_string` is a valid option string.'''

    if not _VALID_OPTION_STRING_RE.fullmatch(option_string):
        return False

    if option_string == '--':
        return False

    return True


def is_valid_variable_name(string):
    '''Check if string is a valid shell variable name.'''

    return _VALID_VARIABLE_RE.fullmatch(string)


def strip_comments(string):
    '''Strip shell comments from string.'''

    lines = []

    for line in string.split('\n'):
        if line.lstrip().startswith('#'):
            continue

        lines.append(line)

    return '\n'.join(lines)


def strip_double_empty_lines(string):
    '''Collapse triple newlines into double newlines.'''

    return string.replace('\n\n\n', '\n\n')


def replace_many(string, replacements):
    '''
    Replace multiple substrings in `string` according to a list
    of (search, replace) tuples.
    '''

    s = string

    for search, replace in replacements:
        s = s.replace(search, replace)

    return s



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for generating a Fish auto completion file.'''

from . import config as config_
from . import utils
from . import algo
from . import fish_complete
from . import fish_helpers
from . import generation
from .output import Output
from .fish_utils import FishCompleteCommand, VariableManager
from .fish_conditions import (
    Conditions, Not,
    HasHiddenOption, HasOption, PositionalNum, PositionalContains
)


class FishCompletionDefinition:
    '''Class holding a completion definition.'''

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    # pylint: disable=too-few-public-methods

    def __init__(
          self,
          ctxt,
          short_options=None,       # List of short options
          long_options=None,        # List of long options
          old_options=None,         # List of old-style options
          positional=None,          # Positional number
          description=None,         # Description
          requires_argument=False,  # Option requires an argument
          keep_order=False,         # Do not sort completion suggestions
          completion_obj=None):

        if short_options is None:
            short_options = []
        if long_options is None:
            long_options = []
        if old_options is None:
            old_options = []

        self.ctxt = ctxt
        self.short_options = short_options
        self.long_options = long_options
        self.old_options = old_options
        self.positional = positional
        self.description = description
        self.requires_argument = requires_argument
        self.keep_order = keep_order
        self.completion_obj = completion_obj
        self.conditions = Conditions()

    def get_complete_cmd(self, ctxt):
        '''Return a `complete` command for current definition.'''

        cmd = FishCompleteCommand()
        cmd.set_command('$prog', raw=True)

        cmd.add_old_options(self.old_options)
        cmd.add_long_options(self.long_options)

        # The Fish manual for complete says that if short option stacking is
        # not desired, we should add the short options as old options.
        if not ctxt.config.option_stacking:
            cmd.add_old_options(self.short_options)
        else:
            cmd.add_short_options(self.short_options)

        if self.description is not None:
            cmd.set_description(self.description)

        if self.requires_argument:
            cmd.flags.add('r')

        if self.keep_order:
            cmd.flags.add('k')

        cmd.parse_args(self.completion_obj.get_args())

        if ctxt.config.fish_fast:
            cmd.set_condition(self.conditions.unsafe_code(ctxt), raw=True)
        else:
            cmd.set_condition(self.conditions.query_code(ctxt), raw=True)

        return cmd


def _get_positional_contains(option):
    commandlines = option.parent.get_parents(include_self=True)
    del commandlines[0]

    conditions = []

    for commandline in commandlines:
        cmds = utils.get_all_command_variations(commandline)
        num = commandline.parent.get_subcommands().get_positional_num()
        conditions.append(PositionalContains(num, cmds))

    return conditions


class FishCompletionGenerator:
    '''Class for generating completions.'''

    # pylint: disable=too-many-instance-attributes

    def __init__(self, ctxt, commandline, parent=None):
        self.ctxt = ctxt
        self.commandline = commandline
        self.parent = parent
        self.children = []

        self.completer = fish_complete.FishCompleter()
        self.subcommands = self.commandline.get_subcommands()
        self.lines = []
        self.conditions = VariableManager('C')
        self.complete_definitions = []

        for option in self.commandline.get_options():
            self.complete_definitions.append(
                self._complete_option(option))

        for positional in self.commandline.get_positionals():
            self.complete_definitions.append(
                self._complete_positional(positional))

        if self.subcommands:
            self.complete_definitions.append(self._complete_subcommands())

            for sub in self.subcommands.subcommands:
                self.children.append(FishCompletionGenerator(ctxt, sub, self))

        for definition in self.complete_definitions:
            if isinstance(definition.completion_obj, fish_complete.FishCompleteCommandArg):
                cmdline = self
                while cmdline:
                    cmdline._fix_command_arg(definition.positional)
                    cmdline = cmdline.parent

        for definition in self.complete_definitions:
            cmd = definition.get_complete_cmd(self.ctxt)

            if not self.ctxt.config.fish_inline_conditions:
                if cmd.condition is not None:
                    cmd.set_condition(self.conditions.add(cmd.condition.string), raw=True)

            self.lines.append(cmd.get())

    def _complete(self, context, *args):
        return self.completer.complete(context, [], *args)

    def _fix_command_arg(self, positional):
        for definition in self.complete_definitions:
            if not (definition.short_options or
                    definition.long_options or
                    definition.old_options):
                continue

            definition.conditions.add(Not(PositionalNum('>=', positional)))

    def visit(self, callback):
        '''Execute callback for generator objects.'''

        callback(self)
        for child in self.children:
            callback(child)

    def get_all(self):
        '''Return all generator objects.'''

        r = []
        self.visit(r.append)
        return r

    def _complete_option(self, option):
        context = self.ctxt.get_option_context(self.commandline, option)

        if option.complete:
            completion_obj = self._complete(context, *option.complete)
        else:
            completion_obj = self._complete(context, 'none')

        definition = FishCompletionDefinition(
            self.ctxt,
            short_options     = option.get_short_option_strings(),
            long_options      = option.get_long_option_strings(),
            old_options       = option.get_old_option_strings(),
            requires_argument = option.has_required_arg(),
            keep_order        = option.nosort,
            description       = option.help,
            completion_obj    = completion_obj)

        not_has_options = []

        final_options = self.commandline.get_final_option_strings()
        not_has_options.extend(final_options)

        conflicting_options = option.get_conflicting_option_strings()
        not_has_options.extend(conflicting_options)

        if not option.repeatable:
            not_has_options.extend(option.option_strings)

        if not_has_options:
            not_has_options = algo.uniq(not_has_options)
            definition.conditions.add(Not(HasOption(not_has_options)))

        if option.hidden:
            definition.conditions.add(HasHiddenOption(option.option_strings))

        # If we don't inherit options, add a condition to the option that
        # ensures that we're in the right (sub)command.
        if not self.commandline.inherit_options and self.subcommands:
            positional = self.subcommands.get_positional_num()
            definition.conditions.add(PositionalNum('==', positional))

        definition.conditions.extend(_get_positional_contains(option))
        definition.conditions.add_when(option.when_parsed)

        return definition

    def _complete_positional(self, positional):
        context = self.ctxt.get_option_context(self.commandline, positional)

        if positional.complete:
            completion_obj = self._complete(context, *positional.complete)
        else:
            completion_obj = self._complete(context, 'none')

        definition = FishCompletionDefinition(
            self.ctxt,
            positional        = positional.get_positional_num(),
            requires_argument = True,
            keep_order        = positional.nosort,
            description       = positional.help,
            completion_obj    = completion_obj
        )

        definition.conditions.add_when(positional.when_parsed)
        definition.conditions.extend(_get_positional_contains(positional))

        positional_num = positional.get_positional_num()
        operator = '>=' if positional.repeatable else '=='
        definition.conditions.add(PositionalNum(operator, positional_num))

        return definition

    def _complete_subcommands(self):
        positional = self.subcommands
        items = positional.get_choices()
        context = self.ctxt.get_option_context(self.commandline, positional)
        completion_obj = self._complete(context, 'choices', items)

        definition = FishCompletionDefinition(
            self.ctxt,
            description = 'Commands',
            completion_obj = completion_obj
        )

        positional_num = positional.get_positional_num()
        definition.conditions.add(PositionalNum('==', positional_num))
        definition.conditions.extend(_get_positional_contains(positional))
        return definition

    def get_command_comment(self):
        '''Return a comment for the current command.'''
        return '# command %s' % self.commandline.get_command_path()

    def get_options_for_query(self):
        '''Return the options for the query command.'''
        return 'set -l opts "%s"' % (
            utils.get_query_option_strings(self.commandline))


def generate_completion(commandline, config=None):
    '''Code for generating a Fish auto completion file.'''

    if config is None:
        config = config_.Config()

    commandline = generation.enhance_commandline(commandline, config)
    helpers = fish_helpers.FishHelpers(config, commandline.prog)
    ctxt = generation.GenerationContext(config, helpers)
    result = FishCompletionGenerator(ctxt, commandline)

    if helpers.is_used('query'):
        types = utils.get_defined_option_types(commandline)
        if types.short:
            ctxt.helpers.use_function('query', 'short_options')
        if types.long:
            ctxt.helpers.use_function('query', 'long_options')
        if types.old:
            ctxt.helpers.use_function('query', 'old_options')

    output = Output(config, helpers)
    output.add_generation_notice()
    output.add_comments()
    output.add_included_files()
    output.add_helper_functions_code()

    with output.add_as_block() as block:
        block.add("set -l prog '%s'" % commandline.prog)

        if helpers.is_used('query'):
            block.add("set -l query '%s'" % helpers.use_function('query'))

        block.add('')
        block.add('# Delete existing completions')
        block.add('complete -c $prog -e')
        block.add('')
        block.add('# Generally disable file completion')
        block.add('complete -c $prog -x')

    if result.commandline.wraps:
        with output.add_as_block() as block:
            block.add('# Wrap command')
            block.add('complete -c $prog -w %s' % result.commandline.wraps)

    for generator in result.get_all():
        if generator.lines:
            with output.add_as_block() as block:
                block.add(generator.get_command_comment())
                block.add(generator.get_options_for_query())
                block.extend(generator.conditions.get_lines())
                block.extend(generator.lines)

    if commandline.aliases:
        with output.add_as_block() as block:
            for alias in commandline.aliases:
                block.add('complete -c %s -e' % alias)
                block.add('complete -c %s -w %s' % (alias, commandline.prog))

    output.add_vim_modeline('fish')

    return output.get()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Built-int Manual.'''

from .errors import CrazyError
from .str_utils import indent, join_with_wrap


# flake8: noqa: E501
# pylint: disable=line-too-long
# pylint: disable=too-many-lines


%COMMANDS%


class TerminalFormatter:
    '''Format string for terminals.'''

    def __init__(self, use_colors, tab_width):
        self.use_colors = use_colors
        self.tab_width = tab_width
        self.tab = ' ' * tab_width

    def bold(self, s):
        '''Make bold string.'''

        if not self.use_colors:
            return s

        return f"\033[1m{s}\033[0m"

    def underline(self, s):
        '''Make underline string.'''

        if not self.use_colors:
            return s

        return f"\033[4m{s}\033[0m"


def _make_command_section(command, formatter):
    r = formatter.bold('COMMAND')
    r += f'\n{formatter.tab}{command["command"]}'
    return r


def _make_description_section(command, formatter):
    r = formatter.bold('DESCRIPTION')
    r += f'\n{formatter.tab}{command["short"]}'

    if command['long']:
        r += '\n\n'

        if formatter.use_colors:
            r += indent(command['long_colored'].strip(), formatter.tab_width)
        else:
            r += indent(command['long'].strip(), formatter.tab_width)

    return r


def _make_example_section(command, formatter):
    r = formatter.bold('EXAMPLE')
    r += '\n'

    if formatter.use_colors:
        r += indent(command['definition_colored'].strip(), formatter.tab_width)
    else:
        r += indent(command['definition'].strip(), formatter.tab_width)

    return r


def _make_output_section(command, formatter):
    r = formatter.bold('OUTPUT')
    r += '\n'
    r += indent(command['output'].strip(), formatter.tab_width)
    return r


def _make_notes_sectiong(notes, formatter):
    if not notes:
        return ''

    r = [formatter.bold('NOTES')]

    for note in notes:
        note = f'{formatter.tab}- {note}'
        r.append(note)

    return '\n'.join(r)


def _make_see_also_section(command, formatter):
    if not command['also']:
        return ''

    r = formatter.bold('SEE ALSO')
    r += '\n'

    for also_cmd, also_desc in command['also'].items():
        r += f'{formatter.tab}{formatter.bold(also_cmd)}: {also_desc}\n'

    return r.rstrip()


def print_help_for_command(name, use_colors):
    '''Print a manual like help for command `name`.'''

    command = None
    for cmd in COMMANDS:
        if cmd['command'] == name:
            command = cmd
            break

    if not command:
        raise CrazyError("Command not found: %s" % name)

    formatter = TerminalFormatter(use_colors, 4)

    notes = list(command['notes'])
    if command['implemented']:
        r = 'This completer is currently only implemented in '
        r += command['implemented'][0]
        if len(command['implemented']) > 1:
            r += ' and '
            r += command['implemented'][1]
        notes.append(r)

    r = [
        _make_command_section(command, formatter),
        _make_description_section(command, formatter),
        _make_example_section(command, formatter),
        _make_output_section(command, formatter),
        _make_notes_sectiong(notes, formatter),
        _make_see_also_section(command, formatter),
    ]

    print()
    print('\n\n'.join(l for l in r if l))
    print()


def print_help_topic(topic, use_colors):
    '''Print help for `topic`.'''

    try:
        print_help_for_command(topic, use_colors)
    except CrazyError:
        commands = [cmd['command'] for cmd in COMMANDS]
        print('Topic not found')
        print('')
        print('Available completers:')
        print(indent(join_with_wrap(' ', '\n', 40, commands), 4))



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Translations.'''

import os
import locale

# flake8: noqa: E501
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring


def normalize_lang(lang_code):
    '''Normalize locale code to short language like `de`.'''

    if not lang_code:
        return 'en'

    return lang_code.split('_')[0].split('.')[0].lower()


def get_lang():
    '''Get system language.'''

    lang = (
        os.environ.get("LANGUAGE") or
        os.environ.get("LC_ALL") or
        os.environ.get("LC_MESSAGES") or
        os.environ.get("LANG")
    )

    if not lang:
        locale.setlocale(locale.LC_ALL, "")
        lang, _ = locale.getlocale()

        if not lang:
            lang = 'en'

    return normalize_lang(lang)


LANG = get_lang()




#!/usr/bin/env python3

'''This script is for checking error messages of cracy-complete.'''

import os
import sys
import yaml

SCRIPT_FILE = os.path.basename(__file__)
SCRIPT_DIR  = os.path.dirname(__file__)
YAML_TESTS_EXPECTED_INFILE  = 'yaml_expected'
YAML_TESTS_EXPECTED_OUTFILE = 'yaml_expected.new'
DICTIONARY_TESTS_EXPECTED_INFILE  = 'dictionary_expected'
DICTIONARY_TESTS_EXPECTED_OUTFILE = 'dictionary_expected.new'

os.environ['LANGUAGE'] = 'en'

# We want to import the development version of cracy-complete,
# not the installed version.
CRAZY_COMPLETE_DIR = os.path.abspath(os.path.join(SCRIPT_DIR, '..', '..'))
sys.path.insert(0, CRAZY_COMPLETE_DIR)

import crazy_complete # noqa: E402

os.chdir(os.path.dirname(os.path.abspath(__file__)))

class TesterBase:
    def __init__(self, expected_file, new_expected_file):
        self.results = []
        self.expected_file = expected_file
        self.new_expected_file = new_expected_file

    def run(self):
        for file in sorted(os.listdir('.')):
            if os.path.isdir(file):
                self.run_subdirectory(file)

        result_content = '\n'.join(self.results)

        with open(self.new_expected_file, 'w', encoding='UTF-8') as fh:
            fh.write(result_content)

        try:
            with open(self.expected_file, 'r', encoding='UTF-8') as fh:
                expected_content = fh.read()
        except Exception as e:
            print(f'Cannot open `{self.expected_file}`: {e}')
            sys.exit(1)

        if expected_content != result_content:
            print(f'{self.expected_file} differs from {self.new_expected_file}.')
            print('Use diff on those files for details')
            sys.exit(1)

    def run_subdirectory(self, directory):
        self.results.append(f'======= Directory "{directory}" =======')

        for file in sorted(os.listdir(directory)):
            full_path = os.path.join(directory, file)
            basename = file.replace('.yaml', '')
            self.run_test(full_path, basename)


class YamlValidatorTester(TesterBase):
    def run_test(self, file, basename):
        print(f'Testing {file}', end=' ... ')

        have_error = 'No error'

        with open(file, 'r', encoding='UTF-8') as fh:
            content = fh.read()

        try:
            parser = crazy_complete.extended_yaml_parser.ExtendedYAMLParser()
            definitions = parser.parse(content)
            crazy_complete.scheme_validator.validate(definitions)
        except crazy_complete.errors.CrazyError as e:
            have_error = str(e)

        self.results.append('%-40s | %s' % (basename, have_error))

        print()


class DictionaryValidatorTester(TesterBase):
    def run_test(self, file, basename):
        print(f'Testing {file}', end=' ... ')

        have_error = 'No error'

        with open(file, 'r', encoding='UTF-8') as fh:
            definition = list(yaml.safe_load_all(fh))

        try:
            cmdline = crazy_complete.dictionary_source.dictionaries_to_commandline(definition)
            crazy_complete.bash.generate_completion(cmdline, None)
        except crazy_complete.errors.CrazyError as e:
            have_error = str(e)

        self.results.append('%-40s | %s' % (basename, have_error))

        print()

testers = [
    YamlValidatorTester(YAML_TESTS_EXPECTED_INFILE, YAML_TESTS_EXPECTED_OUTFILE),
    DictionaryValidatorTester(DICTIONARY_TESTS_EXPECTED_INFILE, DICTIONARY_TESTS_EXPECTED_OUTFILE)
]

for tester in testers:
    tester.run()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for parsing a command line in Bash.'''

from collections import namedtuple

from . import utils
from .str_utils import indent
from .bash_utils import CasePatterns
from .bash_parser_subcommand_code import (
    make_subcommand_switch_code, get_subcommand_path
)

_PARSER_CODE = '''\
POSITIONALS=()
END_OF_OPTIONS=0

local cmd="root" argi arg i char trailing_chars

for ((argi=1; argi < cword; ++argi)); do
  arg="${words_dequoted[argi]}"

  case "$arg" in
    --)
      END_OF_OPTIONS=1
      POSITIONALS+=("${words_dequoted[@]:$((++argi))}")
      return;;
    -?*) # ignore '-'
%LONG_OPTION_CASES%
      for ((i=1; i < ${#arg}; ++i)); do
        char="${arg:$i:1}"
        trailing_chars="${arg:$((i + 1))}"
%SHORT_OPTION_CASES%
      done;;
    *)
      POSITIONALS+=("$arg")
%SUBCOMMAND_SWITCH_CODE%
      ;;
  esac
done

for ((; argi <= cword; ++argi)); do
  case "${words_dequoted[argi]}" in
    -?*);;
    *) POSITIONALS+=("${words_dequoted[argi]}");;
  esac
done'''

_OPT_ISSET = '_OPT_ISSET_'


def generate(commandline, variable_manager):
    '''Generate code for parsing the command line.'''

    commandlines = list(reversed(commandline.get_all_commandlines()))
    subcommand_switch_code = make_subcommand_switch_code(commandline)
    long_option_cases = []
    short_option_cases = []

    for cmdline in commandlines:
        option_cases = _generate_option_cases(cmdline, variable_manager)
        command = get_subcommand_path(cmdline)
        if cmdline.inherit_options:
            command += '*'

        if option_cases.long_options:
            r = 'case "$cmd" in %s)\n' % command
            r += '  case "$arg" in\n'
            for case in option_cases.long_options:
                r += '%s\n' % indent(case, 4)
            r += '  esac\n'
            r += 'esac'
            long_option_cases.append(r)

        if option_cases.short_options:
            r = 'case "$cmd" in %s)\n' % command
            r += '  case "$char" in\n'
            for case in option_cases.short_options:
                r += '%s\n' % indent(case, 4)
            r += '  esac\n'
            r += 'esac'
            short_option_cases.append(r)

    s = _PARSER_CODE

    if long_option_cases:
        s = s.replace('%LONG_OPTION_CASES%',
                      indent('\n\n'.join(long_option_cases), 6))
    else:
        s = s.replace('%LONG_OPTION_CASES%\n', '')

    if short_option_cases:
        s = s.replace('%SHORT_OPTION_CASES%',
                      indent('\n\n'.join(short_option_cases), 8))
    else:
        s = s.replace('%SHORT_OPTION_CASES%\n', '')

    if subcommand_switch_code:
        s = s.replace('%SUBCOMMAND_SWITCH_CODE%',
                      indent(subcommand_switch_code, 6))
    else:
        s = s.replace('%SUBCOMMAND_SWITCH_CODE%\n', '')

    return s


def _make_long_option_case(long_options, option, variable):
    r = ''

    if option.has_optional_arg():
        r += '%s)\n'         % CasePatterns.for_long_without_arg(long_options)
        r += '  %s+=(%s);\n' % (variable, _OPT_ISSET)
        r += '  continue;;\n'
        r += '%s)\n'         % CasePatterns.for_long_with_arg(long_options)
        r += '  %s+=("${arg#*=}")\n' % variable
        r += '  continue;;'
    elif option.has_required_arg():
        r += '%s)\n'         % CasePatterns.for_long_without_arg(long_options)
        r += '  %s+=("${words_dequoted[++argi]}")\n' % variable
        r += '  continue;;\n'
        r += '%s)\n'         % CasePatterns.for_long_with_arg(long_options)
        r += '  %s+=("${arg#*=}")\n'        % variable
        r += '  continue;;'
    else:
        r += '%s)\n'        % CasePatterns.for_long_without_arg(long_options)
        r += '  %s+=(%s)\n' % (variable, _OPT_ISSET)
        r += '  continue;;'

    return r


def _make_short_option_case(short_options, option, variable):
    r = ''

    if option.has_optional_arg():
        r += '%s)\n' % CasePatterns.for_short(short_options)
        r += '  if [[ -n "$trailing_chars" ]]\n'
        r += '  then %s+=("$trailing_chars")\n' % variable
        r += '  else %s+=(%s)\n'                % (variable, _OPT_ISSET)
        r += '  fi\n'
        r += '  continue 2;;'
    elif option.has_required_arg():
        r += '%s)\n' % CasePatterns.for_short(short_options)
        r += '  if [[ -n "$trailing_chars" ]]\n'
        r += '  then %s+=("$trailing_chars")\n'            % variable
        r += '  else %s+=("${words_dequoted[++argi]}")\n'  % variable
        r += '  fi\n'
        r += '  continue 2;;'
    else:
        r += '%s)\n'        % CasePatterns.for_short(short_options)
        r += '  %s+=(%s);;' % (variable, _OPT_ISSET)

    return r


def _generate_option_cases(commandline, variable_manager):
    OptionCases = namedtuple('OptionCases', ['long_options', 'short_options'])
    options = commandline.get_options()
    abbreviations = utils.get_option_abbreviator(commandline)
    option_cases = OptionCases([], [])

    for option in options:
        short_options = option.get_short_option_strings()

        long_options = abbreviations.get_many_abbreviations(
                option.get_long_option_strings())

        long_options += abbreviations.get_many_abbreviations(
                option.get_old_option_strings())

        variable = variable_manager.capture_variable(option)

        if long_options:
            option_cases.long_options.append(
                _make_long_option_case(long_options, option, variable))

        if short_options:
            option_cases.short_options.append(
                _make_short_option_case(short_options, option, variable))

    return option_cases



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Fish utility functions and classes.'''

from .errors import InternalError
from . import shell


class FishString:
    '''A utility class for command-line strings that may require escaping.

    When building command-line commands, it's important to ensure that strings
    are properly escaped to prevent shell injection vulnerabilities or
    syntax errors. However, in some cases, the string may already be escaped,
    and applying escaping again would result in incorrect behavior.
    The `FishString` class provides an abstraction to manage both escaped and
    unescaped strings.

    Attributes:
        string (str):
            The string that represents the command or command argument.

        raw (bool):
            A flag that indicates whether the string is already escaped.
            If `True`, the string is assumed to be already escaped and will
            not be escaped again.

    Args:
        string (str):
            The string to be used in the command.

        raw (bool, optional):
            If `True`, the string will be treated as already escaped.
            Defaults to `False`.

    Methods:
        escape():
            Returns the escaped version of the string, unless the string is
            already escaped (`raw=True`).

        __str__():
            Returns the raw string `string` for direct use or display.
    '''

    def __init__(self, string, raw=False):
        '''Initializes a FishString instance.

        Args:
            string (str):
                The string to be used for command-line purposes.

            raw (bool, optional):
                If True, indicates that the string `s` is already escaped
                and should not be escaped again. Defaults to False.
        '''
        assert isinstance(string, str)
        assert isinstance(raw, bool)
        self.string = string
        self.raw = raw

    def escape(self):
        '''Escapes the string unless it's marked as already escaped.

        If the `raw` attribute is set to `True`, the method returns the string
        without any changes.

        Otherwise, it applies shell escaping to ensure the string is safe to
        use in a command-line context.

        Returns:
            str: The escaped string, or the raw string if it's already escaped.
        '''
        if self.raw:
            return self.string
        return shell.quote(self.string)

    def __str__(self):
        return self.string

    def __repr__(self):
        return repr(self.string)


def make_fish_string(s, raw):
    '''Make a Fish string.'''

    if s is not None:
        return FishString(s, raw)

    return None


class FishCompleteCommand:
    '''Class for creating Fish's `complete` command.'''

    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        self.command       = None
        self.description   = None
        self.condition     = None
        self.arguments     = None
        self.short_options = []
        self.long_options  = []
        self.old_options   = []
        self.flags         = set()

    def set_command(self, command, raw=False):
        '''Set the command (-c|--command).'''
        self.command = FishString(command, raw)

    def set_description(self, description, raw=False):
        '''Set the description (-d|--description).'''
        self.description = make_fish_string(description, raw)

    def add_short_options(self, opts, raw=False):
        '''Add short options (-s|--short-option).'''
        self.short_options.extend(FishString(o.lstrip('-'), raw) for o in opts)

    def add_long_options(self, opts, raw=False):
        '''Add long options (-l|--long-option).'''
        self.long_options.extend(FishString(o.lstrip('-'), raw) for o in opts)

    def add_old_options(self, opts, raw=False):
        '''Add old options (-o|--old-option).'''
        self.old_options.extend(FishString(o.lstrip('-'), raw) for o in opts)

    def set_condition(self, condition, raw=False):
        '''Set the condition (-n|--condition).'''
        self.condition = make_fish_string(condition, raw)

    def set_arguments(self, arguments, raw=False):
        '''Set the arguments (-a|--arguments).'''
        self.arguments = make_fish_string(arguments, raw)

    def add_flag(self, flag):
        '''Set a flag.'''
        self.flags.add(flag)

    def parse_args(self, args):
        '''Parse args (an iterable) and apply it to the instance.'''
        args = list(args)

        while len(args) > 0:
            arg = args.pop(0)

            if arg == '-f':
                self.add_flag('f')
            elif arg == '-F':
                self.add_flag('F')
            elif arg == '-a':
                try:
                    self.set_arguments(args.pop(0))
                except IndexError as e:
                    raise InternalError('Option `-a` requires an argument') from e
            else:
                raise InternalError(f'Unknown option for `complete`: {arg}')

    def get(self):
        '''Return the `complete` command.'''

        r = ['complete']

        if self.command is not None:
            r.extend(['-c', self.command])

        if self.condition is not None:
            r.extend(['-n', self.condition])

        for o in self.short_options:
            r.extend(['-s', o])

        for o in self.long_options:
            r.extend(['-l', o])

        for o in self.old_options:
            r.extend(['-o', o])

        if self.description is not None:
            r.extend(['-d', self.description])

        # -r -f is the same as -x
        if 'r' in self.flags and 'f' in self.flags:
            self.flags.add('x')

        # -x implies -r -f
        if 'x' in self.flags:
            self.flags.discard('r')
            self.flags.discard('f')

        if self.flags:
            r.append('-%s' % ''.join(sorted(self.flags)))

        if self.arguments is not None:
            r.extend(['-a', self.arguments])

        # Pylint doesn't understand that .escape is only called when the
        # value is not a regular Python string.
        # pylint: disable=no-member
        return ' '.join(v if isinstance(v, str) else v.escape() for v in r)


class VariableManager:
    '''
    Manages the creation of unique shell variables for a given name prefix.
    '''

    def __init__(self, variable_name):
        self.variable_name = variable_name
        self.value_to_variable  = {}
        self.counter = 0

    def add(self, value):
        '''Add a value and get its associated shell variable.'''

        if value in self.value_to_variable:
            return '$%s' % self.value_to_variable[value]

        var = '%s%03d' % (self.variable_name, self.counter)
        self.value_to_variable[value] = var
        self.counter += 1
        return '$%s' % var

    def get_lines(self):
        '''Generate shell code to define all stored variables.'''

        r = []
        for value, variable in self.value_to_variable.items():
            r.append('set -l %s %s' % (variable, value))
        return r



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Utility functions for Zsh.'''

from . import algo
from . import shell
from . import string_stream

# pylint: disable=too-many-branches
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments


def escape_colon(s):
    '''Escape colons in a string with backslash.'''

    return s.replace(':', '\\:')


def escape_square_brackets(s):
    '''Escape square brackets with backslash.'''

    return s.replace('[', '\\[').replace(']', '\\]')


def escape_colon_in_quoted_string(s):
    '''Escape a colon in a single/double quoted string.'''

    if ':' not in s:
        return s

    ss = string_stream.StringStream(s)
    result = ''

    while ss.have():
        c = ss.get()
        if c == "'":
            sub = ss.parse_shell_single_quote(in_quotes=True)
            result += "'%s'" % sub.replace(':', r'\:')
        elif c == '"':
            sub = ss.parse_shell_double_quote(in_quotes=True)
            result += '"%s"' % sub.replace(':', r'\\:')
        elif c == ':':
            result += r'\\:'
        else:
            result += c

    return result


def make_option_spec(
        option_strings,
        conflicting_options = None,
        description         = None,
        complete            = None,
        optional_arg        = False,
        repeatable          = False,
        hidden              = False,
        final               = False,
        metavar             = None,
        action              = None,
        long_opt_arg_sep    = 'both'):

    '''
    Make a Zsh option spec.

    Returns something like this:
        (--option -o){--option=,-o+}[Option description]:Metavar:Action
    '''

    result = []

    if conflicting_options is None:
        conflicting_options = []

    # Hidden option ===========================================================
    if hidden:
        result.append("'!'")

    # Not options =============================================================
    not_options = []

    for o in sorted(conflicting_options):
        not_options.append(escape_colon(o))

    if not repeatable:
        for o in sorted(option_strings):
            not_options.append(escape_colon(o))

    if final:
        not_options = ['- *']

    if not_options:
        result.append(shell.quote('(%s)' % ' '.join(algo.uniq(not_options))))

    # Repeatable option =======================================================
    if repeatable:
        result.append("'*'")

    # Option strings ==========================================================
    if complete and optional_arg is True:
        opts = [o+'-' if len(o) == 2 else o+'=-' for o in option_strings]
    elif complete:
        if long_opt_arg_sep == 'both':
            opts = [o+'+' if len(o) == 2 else o+'=' for o in option_strings]
        elif long_opt_arg_sep == 'equals':
            opts = [o+'+' if len(o) == 2 else o+'=-' for o in option_strings]
        elif long_opt_arg_sep == 'space':
            opts = [o+'+' if len(o) == 2 else o for o in option_strings]
    else:
        opts = option_strings

    if len(opts) == 1:
        result.append(opts[0])
    else:
        result.append('{%s}' % ','.join(opts))

    # Description =============================================================
    if description is not None:
        result.append(shell.quote('[%s]' % escape_colon(escape_square_brackets(description))))

    # Complete ================================================================
    if complete:
        if metavar is None:
            metavar = ' '

        action = escape_colon_in_quoted_string(action)
        result.append(':%s:%s' % (shell.quote(escape_colon(metavar)), action))

    return ''.join(result)


def make_positional_spec(
        number,
        repeatable,
        description,
        action):

    '''
    Make a Zsh positional spec.

    Returns something like this:
        Number:Description:Action
    '''

    if repeatable is True:
        num_spec = "'*'"
    else:
        num_spec = str(number)

    desc_spec = shell.quote(escape_colon(description))

    if action == '_normal':
        return f'{num_spec}::{desc_spec}:{action}'

    action_escaped = escape_colon_in_quoted_string(action)
    return f'{num_spec}:{desc_spec}:{action_escaped}'


def make_file_extension_pattern(extensions, fuzzy):
    '''Make a case-insensitive glob pattern matching `extensions`.

    Takes a list of extensions (e.g. ['txt', 'jpg']) and returns a Zsh-style
    pattern that matches all of them, ignoring case.

    Example output:
        '*.(#i)(txt|jpg)'
    '''

    if len(extensions) == 1:
        pattern = '*.(#i)%s' % extensions[0]
    else:
        pattern = '*.(#i)(%s)' % '|'.join(extensions)

    if fuzzy:
        pattern += '*'

    return pattern


def test():
    '''Tests.'''

    cases = [
        (0,  r'foo',         r'foo'),
        (1,  r':',           r'\\:'),
        (2,  r'":"',         r'"\\:"'),
        (3,  r"':'",         r"'\:'"),
    ]

    for num, instring, expected in cases:
        result = escape_colon_in_quoted_string(instring)
        if result != expected:
            print(f"Test {num} failed:")
            print(f"Having:   {result}")
            print(f"Expected: {expected}")
            raise AssertionError("Test failed")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains helper functions for Fish.'''

from .helpers import GeneralHelpers, FishFunction

_QUERY = FishFunction('query', r'''
# ===========================================================================
#
# This function implements the parsing of options and positionals in the
# Fish shell.
#
# Usage: query <OPTIONS> <COMMAND> [ARGS...]
#
# The first argument is a comma-separated list of options that the parser
# should know about.
#
# Short options (-o), long options (--option), and old-style options (-option)
# are supported.
#
# If an option takes an argument, it is suffixed by '='.
# If an option takes an optional argument, it is suffixed by '=?'.
#
# For example:
#   query '-f,--flag,-old-style,--with-arg=,--with-optional=?' [...]
#
#   Here, -f, --flag and -old-style don't take options, --with-arg requires an
#   argument and --with-optional takes an optional argument.
#
# COMMANDS
#   positional_contains <NUM> <WORDS...>
#     Checks if the positional argument number NUM is one of WORDS.
#     NUM counts from one.
#
#   has_option [WITH_INCOMPLETE] <OPTIONS...>
#     Checks if an option given in OPTIONS is passed on commandline.
#     If an option requires an argument, this command returns true only if the
#     option includes an argument. If 'WITH_INCOMPLETE' is specified, it also
#     returns true for options missing their arguments.
#
#   option_is <OPTIONS...> -- <VALUES...>
#     Checks if any option in OPTIONS has a value of VALUES.
#
#   num_of_positionals [<OPERATOR> <NUMBER>]
#     If no arguments are provided, print the count of positional arguments.
#     If two arguments are provided, the first argument should be one of
#     the comparison operators: '-lt', '-le', '-eq', '-ne', '-gt', '-ge'.
#     Returns 0 if the count of positional arguments matches the
#     specified NUMBER according to the comparison operator, otherwise
#     returns 1.
#
# ===========================================================================

set -l positionals
#ifdef positionals_positions
set -l positionals_positions
#endif
set -l having_options
set -l option_values

#ifdef DEBUG
switch (count $argv)
  case 0
    echo '%FUNCNAME%: missing OPTIONS argument' >&2
    return 1
  case 1
    echo '%FUNCNAME%: missing COMMAND' >&2
    return 1
end
#endif

set -l options $argv[1]
set -e argv[1]

set -l cmd $argv[1]
set -e argv[1]

set -l my_cache_key "$(commandline -b) $options"

if test "$__QUERY_CACHE_KEY" = "$my_cache_key"
  set positionals    $__QUERY_CACHE_POSITIONALS
#ifdef positionals_positions
  set positionals_positions $__QUERY_CACHE_POSITIONALS_POSITIONS
#endif
  set having_options $__QUERY_CACHE_HAVING_OPTIONS
  set option_values  $__QUERY_CACHE_OPTION_VALUES
else
  # =========================================================================
  # Parsing of OPTIONS argument
  # =========================================================================

#ifdef short_options
  set -l short_opts_with_arg
  set -l short_opts_without_arg
  set -l short_opts_with_optional_arg
#endif
#ifdef long_options
  set -l long_opts_with_arg
  set -l long_opts_without_arg
  set -l long_opts_with_optional_arg
#endif
#ifdef old_options
  set -l old_opts_with_arg
  set -l old_opts_without_arg
  set -l old_opts_with_optional_arg
#endif

  set -l option

  if test -n "$options"
    for option in (string split -- ' ' $options)
      if false
        true
#ifdef long_options
      else if string match -qr -- '^--.+=$' $option
        set -a long_opts_with_arg (string replace -- '=' '' $option)
      else if string match -qr -- '^--.+=\?$' $option
        set -a long_opts_with_optional_arg (string replace -- '=?' '' $option)
      else if string match -qr -- '^--.+$' $option
        set -a long_opts_without_arg $option
#endif
#ifdef short_options
      else if string match -qr -- '^-.=$' $option
        set -a short_opts_with_arg (string replace -- '=' '' $option)
      else if string match -qr -- '^-.=\?$' $option
        set -a short_opts_with_optional_arg (string replace -- '=?' '' $option)
      else if string match -qr -- '^-.$' $option
        set -a short_opts_without_arg $option
#endif
#ifdef old_options
      else if string match -qr -- '^-..+=$' $option
        set -a old_opts_with_arg (string replace -- '=' '' $option)
      else if string match -qr -- '^-..+=\?$' $option
        set -a old_opts_with_optional_arg (string replace -- '=?' '' $option)
      else if string match -qr -- '^-..+$' $option
        set -a old_opts_without_arg $option
#endif
      end
    end
  end

  # =========================================================================
  # Parsing of options and positionals
  # =========================================================================

  set -l cmdline (commandline -poc)
  set -l cmdline_count (count $cmdline)

  set -l argi 2 # cmdline[1] is command name
  while test $argi -le $cmdline_count
    set -l arg "$cmdline[$argi]"
    set -l have_trailing_arg (test $argi -lt $cmdline_count && echo true || echo false)

    switch $arg
      case '-'
        set -a positionals -
#ifdef positionals_positions
        set -s positionals_positions $argi
#endif
      case '--'
        set -a positionals $cmdline[$(math $argi + 1)..]
#ifdef positionals_positions
        set -a positionals_positions (seq (math $argi + 1) $cmdline_count)
#endif
        break
      case '--*=*'
        set -l split (string split -m 1 -- '=' $arg)
        set -a having_options $split[1]
        set -a option_values "$split[2]"
      case '--*'
#ifdef long_options
        if contains -- $arg $long_opts_with_arg
          if $have_trailing_arg
            set -a having_options $arg
            set -a option_values $cmdline[(math $argi + 1)]
            set argi (math $argi + 1)
          end
        else
          set -a having_options $arg
          set -a option_values ''
        end
#endif
      case '-*'
        set -l end_of_parsing false
#ifdef old_options

        if string match -q -- '*=*' $arg
          set -l split (string split -m 1 -- '=' $arg)
          if contains -- $split[1] $old_opts_with_arg $old_opts_with_optional_arg
            set -a having_options $split[1]
            set -a option_values "$split[2]"
            set end_of_parsing true
          end
        else if contains -- $arg $old_opts_with_arg
          set end_of_parsing true
          if $have_trailing_arg
            set -a having_options $arg
            set -a option_values $cmdline[(math $argi + 1)]
            set argi (math $argi + 1)
          end
        else if contains -- $arg $old_opts_without_arg $old_opts_with_optional_arg
          set -a having_options $arg
          set -a option_values ''
          set end_of_parsing true
        end
#endif
#ifdef short_options

        set -l arg_length (string length -- $arg)
        set -l i 2
        while not $end_of_parsing; and test $i -le $arg_length
          set -l option "-$(string sub -s $i -l 1 -- $arg)"
          set -l trailing_chars "$(string sub -s (math $i + 1) -- $arg)"

          if contains -- $option $short_opts_without_arg
            set -a having_options $option
            set -a option_values ''
          else if contains -- $option $short_opts_with_arg
            set end_of_parsing true

            if test -n "$trailing_chars"
              set -a having_options $option
              set -a option_values $trailing_chars
            else if $have_trailing_arg
              set -a having_options $option
              set -a option_values $cmdline[(math $argi + 1)]
              set argi (math $argi + 1)
            end
          else if contains -- $option $short_opts_with_optional_arg
            set end_of_parsing true
            set -a having_options $option
            set -a option_values "$trailing_chars" # may be empty
          end

          set i (math $i + 1)
        end
#endif
      case '*'
        set -a positionals $arg
#ifdef positionals_positions
        set -a positionals_positions $argi
#endif
    end

    set argi (math $argi + 1)
  end

  set -g __QUERY_CACHE_POSITIONALS    $positionals
#ifdef positionals_positions
  set -g __QUERY_CACHE_POSITIONALS_POSITIONS $positionals_positions
#endif
  set -g __QUERY_CACHE_HAVING_OPTIONS $having_options
  set -g __QUERY_CACHE_OPTION_VALUES  $option_values
  set -g __QUERY_CACHE_KEY            $my_cache_key
end

# ===========================================================================
# Commands
# ===========================================================================

switch $cmd
#ifdef positional_contains
  case 'positional_contains'
#ifdef DEBUG
    if test (count $argv) -eq 0
      echo '%FUNCNAME%: positional_contains: argv[3]: missing number' >&2
      return 1
    end

#endif
    set -l positional_num $argv[1]
    set -e argv[1]
    contains -- $positionals[$positional_num] $argv && return 0 || return 1
#endif
#ifdef has_option
  case 'has_option'
#ifdef with_incomplete
    set -l with_incomplete false

    if test $argv[1] = 'WITH_INCOMPLETE'
      set with_incomplete true
      set -e argv[1]
    end

#endif
    for option in $having_options
      contains -- $option $argv && return 0
    end

#ifdef with_incomplete
    if $with_incomplete
      set -l tokens (commandline -po)
      set -e tokens[1]
      for option in $argv
        if test 2 -eq (string length -- $option)
          set option (string sub -l 1 -s 2 -- $option)
          string match -rq -- "^-[A-z0-9]*$option\$"   $tokens[-2] && return 0
          string match -rq -- "^-[A-z0-9]*$option.*\$" $tokens[-1] && return 0
        else
          string match -q -r -- "^$option\$"       $tokens[-2] && return 0
          string match -q -r -- "^$option(=.*)?\$" $tokens[-1] && return 0
          contains -- $tokens[-1] $argv && return 0
        end
      end
    end

#endif
    return 1
#endif
#ifdef num_of_positionals
  case 'num_of_positionals'
    switch (count $argv)
      case 0
        count $positionals
#ifdef DEBUG
      case 1
        echo '%FUNCNAME%: num_of_positionals: $argv[1]: missing operand' >&2
        return 1
#endif
      case 2
        if contains -- $argv[1] -lt -le -eq -ne -gt -ge;
          test (count $positionals) $argv[1] $argv[2] && return 0 || return 1
#ifdef DEBUG
        else
          echo '%FUNCNAME%: num_of_positionals: $argv[1]: unknown operator' >&2
          return 1
#endif
        end
#ifdef DEBUG
      case '*'
        echo '%FUNCNAME%: num_of_positionals: too many arguments' >&2
        return 1
#endif
    end
#endif
#ifdef option_is
  case 'option_is'
    set -l eof_string (contains -i -- -- $argv || math (count $argv) + 1)
    set -l options $argv[1..$(math $eof_string - 1)]
    set -l values $argv[$eof_string..]

#ifdef DEBUG
    if test (count $options) -eq 0
      echo '%FUNCNAME%: missing options' >&2
      return 1
    end

    if test (count $values) -eq 0
      echo '%FUNCNAME%: missing values' >&2
      return 1
    end

#endif
    set -l i (count $having_options)
    while test $i -ge 1
      if contains -- $having_options[$i] $options
        if contains -- $option_values[$i] $values
          return 0
        end
      end

      set i (math $i - 1)
    end

    return 1
#endif
#ifdef positionals_positions
  case 'positional_pos'
    echo $positionals_positions[$argv[1]]
#endif
#ifdef DEBUG
  case '*'
    echo '%FUNCNAME%: argv[2]: invalid command' >&2
    return 1
#endif
end
''')

_GET_COMPLETING_ARG = FishFunction('get_completing_arg', r'''
set -l arg (commandline -ct | string unescape)

switch $arg
  case '--*=' '--*=*'
    set arg (string replace -r -- '^-[^=]*=' '' $arg)
  case '-*'
    set -l prog (commandline -po)[1]
    set -l progdef (complete -c $prog)

    set -l full_opt (string match -r -- '^-[a-zA-Z0-9]+=' $arg)
    set -l opt (string sub -s 2 -e -1 -- $full_opt)
    set -l optdefs (string match -re -- " -(o|-old-option) $opt( |\$)" $progdef)
    set -l optdefs (string match -re -- " -(x|r|a|-(exclusive|require-parameter|arguments))( |\$)" $optdefs)

    if test (count $optdefs) -gt 0
      set arg (string replace -m 1 -- $full_opt '' $arg)
    else
      set -l i 2

      while test $i -lt (string length -- $arg)
        set -l opt (string sub -s $i -l 1 -- $arg)
        set -l optdefs (string match -re -- " -(s|-short-option) $opt( |\$)" $progdef)
        set -l optdefs (string match -re -- " -(x|r|a|-(exclusive|require-parameter|arguments))( |\$)" $optdefs)

        if test (count $optdefs) -gt 0
          set arg (string sub -s (math $i + 1) -- "$arg")
          break
        end

        set i (math $i + 1)
      end
    end
end

if test -n "$__fish_stripprefix"
  string replace -r -- $__fish_stripprefix '' "$arg"
else
  printf '%s\n' "$arg"
end
''')

_FILEDIR = FishFunction('filedir', r'''
# Function for completing files or directories
#
# Options:
#   -d|--description=DESC   The description for completed entries
#   -D|--directories        Only complete directories
#   -C|--cd=DIR             List contents in DIR
#ifdef regex
#   -r|--regex=PATTERN      Only list files matching pattern
#endif
#ifdef regex_ignore
#   -i|--ignore=PATTERN     Ignore files matching pattern
#endif

argparse --max-args 0 'd/description=' 'D/directories' 'C/cd=' \
#ifdef regex
    'r/regex=' \
#endif
#ifdef regex_ignore
    'i/ignore=' \
#endif
    -- $argv || return 1

set -l comp (get_completing_arg)
set -l desc
set -l files

if set -q _flag_cd[1]
  pushd $_flag_cd 2>/dev/null || return 1
  set files (complete -C"'' $comp")
  popd
else
  set files (complete -C"'' $comp")
end

if set -q _flag_description[1]
  set desc $_flag_description
else if set -g _flag_directories
  set desc 'Directory'
end

if set -q files[1]
  if set -q _flag_directories[1]
    set files (printf '%s\n' $files | string match -r '.*/$')
  end
#ifdef regex

  if set -q _flag_regex[1]
    set files (printf '%s\n' $files | string match -rg "(.*/\$)|($_flag_regex[1])\$")
  end
#endif
#ifdef regex_ignore

  if set -q _flag_ignore[1]
    set files (printf '%s\n' $files | string match -rv "(^|.*/)($_flag_ignore[1])\$")
  end
#endif

  printf '%s\n' $files\t"$desc"
end
''', ['get_completing_arg'])

_LIST = FishFunction('list', r'''
set -l duplicates false

if test $argv[1] = '-d'
  set duplicates true
  set -e argv[1]
end

set -l separator $argv[1]
set -l func $argv[2]
set -l comp (get_completing_arg)

if test -z "$comp"
  eval $func
  return
end

set -l i
set -l value
set -l values
set -l descriptions

set -g __fish_stripprefix "^.*"(string escape --style=regex -- $separator)

for value in (eval $func)
  set -l split (string split -- \t $value)
  set -a values $split[1]
  set -a descriptions "$split[2]"
end

set -e __fish_stripprefix

set -l having_values (string split -- $separator $comp)
set -l remaining_values_idxs

if $duplicates
  set remaining_values_idxs (seq 1 (count $values))
else
  set i 1
  for value in $values
    if not contains -- $value $having_values
      set -a remaining_values_idxs $i
    end

    set i (math $i + 1)
  end
end

switch $comp
  case "*$separator"
    for i in $remaining_values_idxs
      printf '%s%s\t%s\n' $comp $values[$i] "$descriptions[$i]"
    end
  case "*$separator*"
    set comp (string split -r -m 1 -- $separator $comp)[1]

    for i in $remaining_values_idxs
      printf '%s%s%s\t%s\n' "$comp" $separator $values[$i] "$descriptions[$i]"
    end
  case '*'
    for i in $remaining_values_idxs
      printf '%s\t%s\n' $values[$i] "$descriptions[$i]"
    end
end
''', ['get_completing_arg'])

_PREFIX = FishFunction('prefix', r'''
set -l comp (get_completing_arg)
set -l prefix_escaped (string escape --style=regex -- $argv[1])
if string match -qr -- $prefix_escaped $comp
  set -g __fish_stripprefix "^"$prefix_escaped
  printf "%s\n" $argv[1](eval $argv[2])
  set -e __fish_stripprefix
else
  printf "%s\n" $argv[1]
end
''', ['get_completing_arg'])

_KEY_VALUE_LIST = FishFunction('key_value_list', r'''
set -l sep1 $argv[1]
set -l sep2 $argv[2]
set -l value
set -l keys
set -l descriptions
set -l functions
set -l i

for i in (seq 3 3 (count $argv))
  set -a keys $argv[$i]
  set -a descriptions $argv[(math $i + 1)]
  set -a functions $argv[(math $i + 2)]
end

set -l comp (get_completing_arg)

if test -z "$comp" || test (string sub -s -1 -l 1 -- $comp) = $sep1
  for i in (seq 1 (count $keys))
    if test "$functions[$i]" = false
      printf '%s%s\t%s\n' "$comp" $keys[$i] "$descriptions[$i]"
    else
      printf '%s%s%s\t%s\n' "$comp" $keys[$i] $sep2 "$descriptions[$i]"
    end
  end
  return
end

function %PREFIX%__call_func_for_key -S
  set -l i
  for i in (seq 1 (count $keys))
    if test $keys[$i] = $argv[1]
      set -g __fish_stripprefix "^.*"(string escape --style=regex -- $sep2)
      $functions[$i]
      set -e __fish_stripprefix
      return
    end
  end
end

set -l pair (string split -- $sep1 $comp)[-1]
set -l split (string split -- $sep2 $pair)

switch $pair
  case "*$sep2*"
    set -l value_len (string length -- $split[2])

    if test $value_len -gt 0
      set comp (string sub -e -$value_len -- $comp)
    end

    for value in (%PREFIX%__call_func_for_key $split[1])
      printf '%s%s\n' $comp $value
    end
  case '*'
    set -l key_len (string length -- $split[1])
    set comp (string sub -e -$key_len -- $comp)

    for i in (seq 1 (count $keys))
      if test "$functions[$i]" = false
        printf '%s%s\t%s\n' "$comp" $keys[$i] "$descriptions[$i]"
      else
        printf '%s%s%s\t%s\n' "$comp" $keys[$i] $sep2 "$descriptions[$i]"
      end
    end
end
''', ['get_completing_arg'])

_HISTORY = FishFunction('history', r'''
builtin history | command grep -E -o -- $argv[1]
''')

_MIME_FILE = FishFunction('mime_file', r'''
set -l i_opt

set -l comp (get_completing_arg)

if command file -i /dev/null &>/dev/null
  set i_opt '-i'
else if command file -I /dev/null &>/dev/null
  set i_opt '-I'
else
  complete -C"'' $comp"
  return
end

set -l line
command file -L $i_opt -- "$comp"* 2>/dev/null | while read line
  set -l split (string split -m 1 -r ':' "$line")

  if string match -q -- '*inode/directory*' $split[2]
    printf '%s/\n' "$split[1]"
  else if begin; echo "$split[2]" | command grep -q -E -- $argv[1]; end
    printf '%s\n' "$split[1]"
  end
end
''', ['get_completing_arg'])

_SUBSTRACT_PREFIX_SUFFIX = FishFunction('subtract_prefix_suffix', r'''
set -l s1 $argv[1]
set -l s2 $argv[2]

set -l len1 (string length -- $s1)
set -l len2 (string length -- $s2)

set -l maxk $len1
if test $len2 -lt $maxk
  set maxk $len2
end

set -l k $maxk
while test $k -gt 0
  set -l start (math $len1 - $k + 1)
  set -l suf (string sub -s $start -l $k -- $s1)
  set -l pre (string sub -s 1 -l $k -- $s2)

  if test "$suf" = "$pre"
    if test $len1 -eq $k
      echo ''
    else
      string sub -l (math $len1 - $k) -- $s1
    end
    return
  end

  set k (math $k - 1)
end

echo $s1
''')

_COMMANDLINE_STRING = FishFunction('commandline_string', r'''
set -l line
set -l comp (get_completing_arg)

complete -C "$comp" | while read line
  set -l split (string split -m 1 -- \t $line)
  set -l comp2 (subtract_prefix_suffix "$comp" "$split[1]")
  printf '%s\t%s\n' "$comp2$split[1]" "$split[2]"
end''', ['subtract_prefix_suffix', 'get_completing_arg'])

_DATE_FORMAT = FishFunction('date_format', r'''
set -l comp (get_completing_arg)

if test "$(string sub -s -1 -l 1 -- $comp)" = '%'
  printf '%s%s\t%s\n' \
    "$comp" 'a' 'abbreviated day name' \
    "$comp" 'A' 'full day name' \
    "$comp" 'b' 'abbreviated month name' \
    "$comp" 'h' 'abbreviated month name' \
    "$comp" 'B' 'full month name' \
    "$comp" 'c' 'preferred locale date and time' \
    "$comp" 'C' '2-digit century' \
    "$comp" 'd' 'day of month (01-31)' \
    "$comp" 'D' 'American format month/day/year (%m/%d/%y)' \
    "$comp" 'e' 'day of month ( 1-31)' \
    "$comp" 'F' 'ISO 8601 year-month-date (%Y-%m-%d)' \
    "$comp" 'G' '4-digit ISO 8601 week-based year' \
    "$comp" 'g' '2-digit ISO 8601 week-based year' \
    "$comp" 'H' 'hour (00-23)' \
    "$comp" 'I' 'hour (01-12)' \
    "$comp" 'j' 'day of year (001-366)' \
    "$comp" 'k' 'hour ( 0-23)' \
    "$comp" 'l' 'hour ( 1-12)' \
    "$comp" 'm' 'month (01-12)' \
    "$comp" 'M' 'minute (00-59)' \
    "$comp" 'n' 'newline' \
    "$comp" 'p' 'locale dependent AM/PM' \
    "$comp" 'r' 'locale dependent a.m. or p.m. time (%I:%M:%S %p)' \
    "$comp" 'R' '24-hour notation time (%H:%M)' \
    "$comp" 's' 'seconds since the epoch' \
    "$comp" 'S' 'seconds (00-60)' \
    "$comp" 't' 'tab' \
    "$comp" 'T' '24-hour notation with seconds (%H:%M:%S)' \
    "$comp" 'u' 'day of week (1-7, 1=Monday)' \
    "$comp" 'U' 'week number of current year, Sunday based (00-53)' \
    "$comp" 'V' 'ISO 8601 week number of current year, week 1 has 4 days in current year (01-53)' \
    "$comp" 'w' 'day of week (0-6, 0=Sunday)' \
    "$comp" 'W' 'week number of current year, Monday based (00-53)' \
    "$comp" 'x' 'locale dependent date representation without time' \
    "$comp" 'X' 'locale dependent time representation without date' \
    "$comp" 'y' '2-digit year (00-99)' \
    "$comp" 'Y' 'full year' \
    "$comp" 'z' 'UTC offset' \
    "$comp" 'Z' 'timezone name' \
    "$comp" '%' 'literal %'
end
''', ['get_completing_arg'])

_NUMBER = FishFunction('number', r'''
set -l comp (get_completing_arg)
set -l number (string match -r -- '^[0-9,.]*' $comp)

if test -z "$number"
  return
end

set -l suffix_and_desc
for suffix_and_desc in $argv
  set suffix_and_desc (string split -m 1 -- ':' $suffix_and_desc)
  printf '%s%s\t%s\n' $number $suffix_and_desc[1] $suffix_and_desc[2]
end
''', ['get_completing_arg'])

# =============================================================================
# Bonus
# =============================================================================

_TIMEZONE_LIST = FishFunction('timezone_list', r'''
if ! command timedatectl list-timezones 2>/dev/null
  command find /usr/share/zoneinfo -type f |\
  command sed 's|/usr/share/zoneinfo/||g'  |\
  command grep -E -v '^(posix|right)'
end
''')

_ALSA_LIST_CARDS = FishFunction('alsa_list_cards', r'''
set -l card

for card in (command aplay -l | string match -r '^card [0-9]+: [^,]+')
  set card (string replace 'card ' '' $card)
  set -l split (string split ': ' $card)

  if set -q split[2]
    printf "%s\t%s\n" $split[1] $split[2]
  end
end
''')

_ALSA_LIST_DEVICES = FishFunction('alsa_list_devices', r'''
set -l card

for card in (command aplay -l | string match -r '^card [0-9]+: [^,]+')
  set card (string replace 'card ' '' $card)
  set -l split (string split ': ' $card)

  if set -q split[2]
    printf "hw:%s\t%s\n" $split[1] $split[2]
  end
end
''')


class FishHelpers(GeneralHelpers):
    '''Class holding helper functions for Fish.'''

    def __init__(self, config, function_prefix):
        super().__init__(config, function_prefix, FishFunction)
        self.add_function(_QUERY)
        self.add_function(_GET_COMPLETING_ARG)
        self.add_function(_FILEDIR)
        self.add_function(_LIST)
        self.add_function(_PREFIX)
        self.add_function(_KEY_VALUE_LIST)
        self.add_function(_HISTORY)
        self.add_function(_DATE_FORMAT)
        self.add_function(_NUMBER)
        self.add_function(_MIME_FILE)
        self.add_function(_SUBSTRACT_PREFIX_SUFFIX)
        self.add_function(_COMMANDLINE_STRING)
        self.add_function(_TIMEZONE_LIST)
        self.add_function(_ALSA_LIST_CARDS)
        self.add_function(_ALSA_LIST_DEVICES)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Classes for generating file output.'''


_GENERATION_NOTICE = '''\
# This script was generated by crazy-complete.
# crazy-complete: A tool that creates robust and reliable autocompletion scripts for Bash, Fish and Zsh.
# For more information, visit: https://github.com/crazy-complete/crazy-complete'''


class Output:
    '''Class for generating output.'''

    def __init__(self, config, helpers):
        self.config = config
        self.helpers = helpers
        self.output = []

    def add(self, string):
        '''Add to output.'''
        self.output.append(string)

    def extend(self, strings):
        '''Add many to output.'''
        self.output.extend(strings)

    def add_as_block(self):
        '''Return a helper context manager for adding a multi-line block.'''

        class _Block:
            def __init__(self, parent):
                self.parent = parent
                self.lines = []

            def add(self, line):
                '''Add a single line to the current block.'''
                self.lines.append(line)

            def extend(self, lines):
                '''Add multiple lines to current block.'''
                self.lines.extend(lines)

            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                block_text = "\n".join(self.lines)
                self.parent.output.append(block_text)

        return _Block(self)

    def add_generation_notice(self):
        '''Add the generation notice.'''
        self.add(_GENERATION_NOTICE)

    def add_comments(self):
        '''Add additioanl comments.'''
        if self.config.comments:
            self.add('\n'.join(f'# {c}' for c in self.config.comments))

    def add_included_files(self):
        '''Add included files.'''
        for file in self.config.include_files:
            with open(file, 'r', encoding='utf-8') as fh:
                self.add(fh.read().strip())

    def add_helper_functions_code(self):
        '''Add helper functions.'''
        for code in self.helpers.get_used_functions_code():
            self.add(code)

    def add_vim_modeline(self, shell):
        '''Add the vim modeline.'''
        if self.config.vim_modeline:
            vim = 'vim'
            self.add(f'# {vim}: ft={shell} ts=2 sts=2 sw=2 et')

    def get(self):
        '''Return the output.'''
        return '\n\n'.join(filter(None, self.output))



import os
import pyte
import pexpect

from terminal_base import TerminalBase

class PyteTerminal(TerminalBase):
    def __init__(self, width=80, height=100):
        self.width = width
        self.height = height
        self.screen = pyte.Screen(width, height)
        self.stream = pyte.Stream(self.screen)
    
    def start(self, command_args, env=None):
        spawn_env = None
        if env:
            spawn_env = os.environ.copy()
            spawn_env.update(env)

        command = command_args.pop(0)
        self.terminal = pexpect.spawn(command, command_args, dimensions=(self.height, self.width), env=spawn_env)

    def stop(self):
        self.terminal.close()

    def resize_window(self, x, y):
        pass # TODO
    
    def send(self, text):
        self.terminal.send(text)

    def send_line(self, line):
        self.terminal.sendline(line)
    
    def send_tab(self):
        self.terminal.send("\t")
    
    def send_ctrl(self, key):
        self.terminal.sendcontrol(key)
    
    def _read_output(self):
        try:
            while True:
                data = self.terminal.read_nonblocking(size=1024, timeout=0.1)
                self.stream.feed(data.decode('utf-8'))
        except pexpect.exceptions.TIMEOUT:
            pass
    
    def get_output(self):
        self._read_output()
        return "\n".join(self.screen.display)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for creating when conditions in Bash.'''

from . import when
from . import shell


class ConditionGenerator:
    '''Class for generating conditions.'''

    def __init__(self, commandline, variable_manager):
        self.commandline = commandline
        self.variable_manager = variable_manager

    def generate(self, tokens):
        '''Turn tokens/objects into condition code.'''

        r = []

        for obj in tokens:
            if isinstance(obj, when.OptionIs):
                r.append(self._gen_option_is(obj))

            elif isinstance(obj, when.HasOption):
                r.append(self._gen_has_option(obj))

            elif obj in ('&&', '||', '!'):
                r.append(obj)

            elif obj == '(':
                r.append('{')

            elif obj == ')':
                r.append(';}')

            else:
                raise AssertionError("Not reached")

        if when.needs_braces(tokens):
            return '{ %s; }' % ' '.join(r)

        return ' '.join(r)

    def _gen_option_is(self, obj):
        conditions = []

        for o in self.commandline.get_options_by_option_strings(obj.options):
            have_option = '(( ${#%s[@]} ))' % self.variable_manager.capture_variable(o)
            value_equals = []

            for value in obj.values:
                value_equals.append('[[ "${%s[-1]}" == %s ]]' % (
                    self.variable_manager.capture_variable(o),
                    shell.quote(value)
                ))

            if len(value_equals) == 1:
                cond = '{ %s && %s; }' % (have_option, value_equals[0])
            else:
                cond = '{ %s && { %s; } }' % (have_option, ' || '.join(value_equals))

            conditions.append(cond)

        if len(conditions) == 1:
            return conditions[0]

        return '{ %s; }' % ' || '.join(conditions)

    def _gen_has_option(self, obj):
        conditions = []

        for o in self.commandline.get_options_by_option_strings(obj.options):
            cond = '(( ${#%s[@]} ))' % self.variable_manager.capture_variable(o)
            conditions.append(cond)

        if len(conditions) == 1:
            return conditions[0]

        return '{ %s; }' % ' || '.join(conditions)


def generate_when_conditions(commandline, variable_manager, when_):
    '''Generate when condition code.'''

    tokens = when.parse_when(when_)
    generator = ConditionGenerator(commandline, variable_manager)
    return generator.generate(tokens)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Translations.'''

import os
import locale

# flake8: noqa: E501
# pylint: disable=line-too-long
# pylint: disable=missing-function-docstring


def normalize_lang(lang_code):
    '''Normalize locale code to short language like `de`.'''

    if not lang_code:
        return 'en'

    return lang_code.split('_')[0].split('.')[0].lower()


def get_lang():
    '''Get system language.'''

    lang = (
        os.environ.get("LANGUAGE") or
        os.environ.get("LC_ALL") or
        os.environ.get("LC_MESSAGES") or
        os.environ.get("LANG")
    )

    if not lang:
        locale.setlocale(locale.LC_ALL, "")
        lang, _ = locale.getlocale()

        if not lang:
            lang = 'en'

    return normalize_lang(lang)


LANG = get_lang()


def command_arg_without_command():
    if LANG == 'de':
        return 'Vervollständiger `command_arg` benötigt einen vorher definierten `command` Vervollständiger'
    return 'Completer `command_arg` requires a previously defined `command` completer'


def completer_not_allowed_in(completer1, completer2):
    if LANG == 'de':
        return f'Vervollständiger `{completer1}` nicht in `{completer2}` erlaubt'
    return f'Completer `{completer1}` not allowed in `{completer2}`'


def completer_not_allowed_in_option(completer):
    if LANG == 'de':
        return f'Vervollständiger `{completer}` nicht erlaubt in einer Option'
    return f'Completer `{completer}` not allowed inside an option'


def completer_requires_repeatable(completer):
    if LANG == 'de':
        return f'Vervollständiger `{completer}` benötigt `repeatable=True`'
    return f'Completer `{completer}` requires `repeatable=true`'


def dict_cannot_be_empty():
    if LANG == 'de':
        return 'Dictionary darf nicht leer sein'
    return 'Dictionary cannot be empty'


def integer_cannot_be_zero():
    if LANG == 'de':
        return 'Ganzzahl darf nicht Null sein'
    return 'Integer cannot be zero'


def integer_must_be_greater_than_zero():
    if LANG == 'de':
        return 'Ganzzahl muss größer als Null sein'
    return 'Integer must be greater than zero'


def invalid_type_expected_types(types):
    if LANG == 'de':
        return f'Ungültiger Typ. Erwartet: {types}'
    return f'Invalid type. Expected: {types}'


def invalid_value():
    if LANG == 'de':
        return 'Ungültiger Wert'
    return 'Invalid value'


def invalid_value_expected_values(values):
    if LANG == 'de':
        return f'Ungültiger Wert. Erwartet: {values}'
    return f'Invalid value. Expected: {values}'


def list_cannot_be_empty():
    if LANG == 'de':
        return 'Liste darf nicht leer sein'
    return 'List cannot be empty'


def list_must_contain_at_least_two_items():
    if LANG == 'de':
        return 'Liste muss mindestens zwei Elemente enthalten'
    return 'List must contain at least two items'


def list_must_contain_exact_three_items():
    if LANG == 'de':
        return 'Liste muss exakt 3 Elemente enthalten'
    return 'List must contain exactly three items'


def missing_arg():
    if LANG == 'de':
        return 'Fehlendes benötigtes Argument'
    return 'Missing required argument'


def missing_definition_of_program(program):
    if LANG == 'de':
        return f'Fehlende Definition des Programms `{program}`'
    return f'Missing definition of program `{program}`'


def multiple_definition_of_program(program):
    if LANG == 'de':
        return f'Mehrfache Definition des Programms `{program}`'
    return f'Multiple definition of program `{program}`'


def mutually_exclusive_parameters(parameters):
    if LANG == 'de':
        return f'Parameter schließen sich gegenseitig aus: {parameters}'
    return f'Parameters are mutually exclusive: {parameters}'


def no_programs_defined():
    if LANG == 'de':
        return 'Keine Programme definiert'
    return 'No programs defined'


def not_a_variable_name():
    if LANG == 'de':
        return 'Ungültiger Variablenname'
    return 'Not a valid variable name'


def not_an_absolute_path():
    if LANG == 'de':
        return 'Kein absoluter Pfad'
    return 'Not an absolute path'


def not_an_extended_regex():
    if LANG == 'de':
        return 'Ungültiger erweiterter regulärer Ausdruck'
    return 'Not an extended regular expression'


def parameter_not_allowed_in_subcommand(parameter):
    if LANG == 'de':
        return f'Parameter `{parameter}` ist nicht erlaub einem Unterkommando'
    return f'Parameter `{parameter}` not allowed in subcommand'


def parameter_requires_parameter(parameter1, parameter2):
    if LANG == 'de':
        return f'Parameter `{parameter1}` benötigt `{parameter2}`'
    return f'Parameter `{parameter1}` requires `{parameter2}`'


def positional_argument_after_repeatable():
    if LANG == 'de':
        return 'Positionsargument nach einem wiederholbaren Positionsargument gefunden'
    return 'Positional argument found after a repeatable positional argument'


def repeatable_with_subcommands():
    if LANG == 'de':
        return 'Wiederholbare Positionsargumente dürfen nicht zusammen mit Unterkommandos definiert werden'
    return 'Repeatable positional arguments cannot be used together with subcommands'


def single_character_expected():
    if LANG == 'de':
        return 'Ungültige Länge. Einzelner Buchstabe erwartet'
    return 'Invalid length. Single character expected'


def string_cannot_be_empty():
    if LANG == 'de':
        return 'Zeichenkette darf nicht leer sein'
    return 'String cannot be empty'


def string_cannot_contain_space():
    if LANG == 'de':
        return 'Zeichenkette darf keine Leerzeichen enthalten'
    return 'String cannot contain spaces'


def too_many_arguments():
    if LANG == 'de':
        return 'Zu viele Argumente'
    return 'Too many arguments'


def too_many_programs_defined():
    if LANG == 'de':
        return 'Zu viele Programme definiert'
    return 'Too many programs defined'


def too_many_repeatable_positionals():
    if LANG == 'de':
        return 'Zu viele wiederholbare Positionsargumente'
    return 'Too many repeatable positionals'


def unknown_completer():
    if LANG == 'de':
        return 'Unbekannter Vervollständiger'
    return 'Unknown completer'


def unknown_parameter():
    if LANG == 'de':
        return 'Unbekannter Parameter'
    return 'Unknown parameter'



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for generating a Bash auto completion file.'''

from collections import OrderedDict

from . import config as config_
from . import shell
from . import algo
from . import utils
from . import helpers
from . import bash_helpers
from . import bash_complete
from . import bash_parser
from . import bash_parser_v2
from . import bash_option_completion
from . import bash_option_strings_completion
from . import bash_positionals_completion
from . import bash_versions
from . import bash_patterns
from .str_utils import indent, join_with_wrap
from .bash_utils import VariableManager
from .output import Output
from . import generation


class BashCompletionGenerator:
    '''Class for generating completions.'''

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __init__(self, ctxt, commandline):
        self.ctxt = ctxt
        self.commandline = commandline
        self.options = commandline.get_options()
        self.positionals = commandline.get_positionals()
        self.subcommands = commandline.get_subcommands()
        self.completer = bash_complete.BashCompleter()
        self.variable_manager = VariableManager('OPT_')
        self._generate()

    def complete_option(self, option):
        '''Complete an option.'''
        context = self.ctxt.get_option_context(self.commandline, option)
        obj = self.completer.complete(context, [], *option.complete)
        code = obj.get_code(False)
        if code and option.nosort:
            code += '\ncompopt -o nosort'
        return code

    def _generate_commandline_parsing(self):
        local_vars = []
        for cmdline in self.commandline.get_all_commandlines():
            for option in cmdline.options:
                if option.capture is not None:
                    local_vars.append(option.capture)

        r = 'local END_OF_OPTIONS POSITIONALS\n'

        if local_vars:
            local_vars = algo.uniq(local_vars)
            line_length = self.ctxt.config.line_length - 2
            r += join_with_wrap(' ', '\n', line_length, local_vars, 'local -a ')

        r += '\n%s' % self.ctxt.helpers.use_function('parse_commandline')
        return r

    def _generate_subcommand_call(self):
        if not self.subcommands:
            return None

        positional_num = self.subcommands.get_positional_num()
        r  = 'if (( %i < ${#POSITIONALS[@]} )); then\n' % (positional_num - 1)
        r += '  case "${POSITIONALS[%i]}" in\n' % (positional_num - 1)
        for subcommand in self.subcommands.subcommands:
            cmds = utils.get_all_command_variations(subcommand)
            pattern = bash_patterns.make_pattern([shell.quote(s) for s in cmds])
            funcname = self.ctxt.helpers.make_completion_funcname(subcommand)
            if utils.is_worth_a_function(subcommand):
                if self.commandline.inherit_options:
                    r += '    %s) %s && return 0;;\n' % (pattern, funcname)
                else:
                    r += '    %s) %s && return 0 || return 1;;\n' % (pattern, funcname)
            else:
                if self.commandline.inherit_options:
                    r += '    %s);;\n' % pattern
                else:
                    r += '    %s) return 0;;\n' % pattern
        r += '  esac\n'
        r += 'fi'
        return r

    def _generate_command_arg_call(self):
        def is_command_arg(positional):
            return (positional.complete and
                    positional.complete[0] == 'command_arg')

        try:
            positional = list(filter(is_command_arg, self.positionals))[0]
        except IndexError:
            return None

        num = positional.get_positional_num()
        r  = 'if (( ${#POSITIONALS[@]} >= %d )); then\n' % num
        r += '  local realpos\n'
        r += '  for realpos in "${!COMP_WORDS[@]}"; do\n'
        r += '    [[ "${COMP_WORDS[realpos]}" == "${POSITIONALS[%d]}" ]] && {\n' % (num - 2)
        r += '      %s $realpos\n' % bash_versions.command_offset(self.ctxt)
        r += '      return 0;\n'
        r += '    }\n'
        r += '  done\n'
        r += 'fi'
        return r

    def _generate(self):
        # The completion function returns 0 (success) if there was a completion match.
        # This return code is used for dealing with subcommands.

        if not utils.is_worth_a_function(self.commandline):
            self.result = ''
            return

        code = OrderedDict()
        code['init_completion'] = None
        code['command_line_parsing'] = None
        code['subcommand_call'] = None
        code['command_arg'] = None
        code['option_completion'] = None
        code['option_strings_completion'] = None
        code['positional_completion'] = None

        code['subcommand_call'] = self._generate_subcommand_call()
        code['command_arg'] = self._generate_command_arg_call()
        code['option_completion'] = bash_option_completion.generate_option_completion(self)
        code['option_strings_completion'] = bash_option_strings_completion.generate(self)
        code['positional_completion'] = bash_positionals_completion.generate(self)

        if self.commandline.parent is None:
            # The root parser makes those variables local and sets up the completion.
            r  = 'local cur prev words cword split words_dequoted\n'
            r += '%s -n =: || return\n' % bash_versions.init_completion(self.ctxt)
            r += self.ctxt.helpers.use_function('dequote_words')
            code['init_completion'] = r

            v1 = bash_parser.generate(self.commandline, self.variable_manager)
            v2 = bash_parser_v2.generate(self.commandline, self.variable_manager)
            c  = v1 if len(v1) < len(v2) else v2

            func = helpers.ShellFunction('parse_commandline', c)
            self.ctxt.helpers.add_function(func)

            # This sets up END_OF_OPTIONS, POSITIONALS and the OPT_* variables.
            code['command_line_parsing'] = self._generate_commandline_parsing()

        r  = '%s() {\n' % self.ctxt.helpers.make_completion_funcname(self.commandline)
        r += '%s\n\n'   % indent('\n\n'.join(c for c in code.values() if c), 2)
        r += '  return 1\n'
        r += '}'

        self.result = r


def _generate_wrapper(ctxt, commandline):
    make_completion_funcname = ctxt.helpers.make_completion_funcname
    completion_funcname = make_completion_funcname(commandline)
    wrapper_funcname = make_completion_funcname(commandline, '__wrapper')

    if not commandline.wraps:
        return (completion_funcname, None)

    r  = '%s %s\n' % (bash_versions.completion_loader(ctxt), commandline.wraps)
    r += '\n'
    r += '%s() {\n' % wrapper_funcname
    r += '  local WRAPS=%s\n' % shell.quote(commandline.wraps)
    r += '  local COMP_LINE_OLD="$COMP_LINE"\n'
    r += '  local COMP_POINT_OLD="$COMP_POINT"\n'
    r += '  local COMP_WORDS_OLD=("${COMP_WORDS[@]}")\n'
    r += '\n'
    r += '  COMP_LINE="${COMP_LINE/$1/$WRAPS}"\n'
    r += '  COMP_WORDS[0]=%s\n' % shell.quote(commandline.wraps)
    r += '  COMP_POINT=$(( COMP_POINT + ${#WRAPS} - ${#1} ))\n'
    r += '\n'
    r += '  %s 0\n' % bash_versions.command_offset(ctxt)
    r += '\n'
    r += '  COMP_LINE="$COMP_LINE_OLD"\n'
    r += '  COMP_POINT="$COMP_POINT_OLD"\n'
    r += '  COMP_WORDS=("${COMP_WORDS_OLD[@]}")\n'
    r += '  local COMPREPLY_OLD=("${COMPREPLY[@]}")\n'
    r += '  %s "$@"\n' % completion_funcname
    r += '\n'
    r += '  COMPREPLY=("${COMPREPLY_OLD[@]}" "${COMPREPLY[@]}")\n'
    r += '}'

    return (wrapper_funcname, r)


def generate_completion(commandline, config=None):
    '''Code for generating a Bash auto completion file.'''

    if config is None:
        config = config_.Config()

    commandline = generation.enhance_commandline(commandline, config)
    helpers = bash_helpers.BashHelpers(config, commandline.prog)
    ctxt = generation.GenerationContext(config, helpers)

    if ctxt.config.bash_completions_version >= (2, 12):
        helpers.define('bash_completions_v_2_12')

    result = generation.visit_commandlines(BashCompletionGenerator, ctxt, commandline)

    completion_func, wrapper_code = _generate_wrapper(ctxt, commandline)

    output = Output(config, helpers)
    output.add_generation_notice()
    output.add_comments()
    output.add_included_files()
    output.add_helper_functions_code()
    output.extend(generator.result for generator in result)
    output.add(wrapper_code)
    output.add('complete -F %s %s' % (
        completion_func, ' '.join([commandline.prog] + commandline.aliases)
    ))
    output.add_vim_modeline('sh')

    return output.get()
